import numpy as np
import pytest

import tiledb

from .common import DiskTestCase, assert_captured


class FilestoreTest(DiskTestCase):
    @pytest.fixture
    def text_fname(self):
        path = self.path("text_fname")
        vfs = tiledb.VFS()
        vfs.touch(path)
        with vfs.open(path, "wb") as fio:
            fio.write(b"Simple text file.\n")
            fio.write(b"With two lines.")
        return path

    def test_buffer(self, capfd):
        path = self.path("test_buffer")
        data = b"buffer"

        fs = tiledb.Filestore(path)

        with self.assertRaises(tiledb.TileDBError):
            fs.write(data)

        schema = tiledb.ArraySchema.from_file()
        tiledb.Array.create(path, schema)

        assert schema.attr(0).name == "contents"
        assert schema.attr(0).dtype == np.bytes_

        schema.attr(0).dump()
        assert_captured(capfd, "Type: BLOB")

        fs = tiledb.Filestore(path)
        fs.write(data)
        assert bytes(data) == fs.read()

    def test_small_buffer(self, capfd):
        path = self.path("test_small_buffer")
        # create a 4 byte array
        data = b"abcd"

        fs = tiledb.Filestore(path)

        with self.assertRaises(tiledb.TileDBError):
            fs.write(data)

        schema = tiledb.ArraySchema.from_file()
        tiledb.Array.create(path, schema)

        assert schema.attr(0).name == "contents"
        assert schema.attr(0).dtype == np.bytes_

        schema.attr(0).dump()
        assert_captured(capfd, "Type: BLOB")

        fs = tiledb.Filestore(path)
        fs.write(data)
        assert data[3:4] == fs.read(offset=3, size=1)

    def test_uri(self, text_fname):
        path = self.path("test_uri")
        schema = tiledb.ArraySchema.from_file(text_fname)
        tiledb.Array.create(path, schema)

        fs = tiledb.Filestore(path)
        tiledb.Filestore.copy_from(path, text_fname)
        with open(text_fname, "rb") as text:
            data = text.read()
            assert data == fs.read(0, len(data))
            assert len(fs) == len(data)

    def test_multiple_writes(self):
        path = self.path("test_buffer")
        schema = tiledb.ArraySchema.from_file()
        tiledb.Array.create(path, schema)

        fs = tiledb.Filestore(path)
        for i in range(1, 4):
            fs.write(("x" * i).encode())

        assert fs.read() == ("x" * i).encode()

        timestamps = [t[0] for t in tiledb.array_fragments(path).timestamp_range]
        for i, ts in enumerate(timestamps, start=1):
            with tiledb.open(path, timestamp=ts) as A:
                assert A.meta["file_size"] == i



import xml.etree.ElementTree

import numpy as np
import pytest

import tiledb

from .common import DiskTestCase, assert_captured


class DomainTest(DiskTestCase):
    def test_domain(self, capfd):
        dims = [
            tiledb.Dim("d1", (1, 4), 2, dtype="u8"),
            tiledb.Dim("d2", (1, 4), 2, dtype="u8"),
        ]
        dom = tiledb.Domain(*dims)

        # check that dumping works
        dom.dump()
        assert_captured(capfd, "Name: d1")

        self.assertEqual(dom.ndim, 2)
        self.assertEqual(dom.dtype, np.dtype("uint64"))
        self.assertEqual(dom.shape, (4, 4))

        # check that we can iterate over the dimensions
        dim_names = [dim.name for dim in dom]
        self.assertEqual(["d1", "d2"], dim_names)

        # check that we can access dim by name
        dim_d1 = dom.dim("d1")
        self.assertEqual(dim_d1, dom.dim(0))

        # check that we can construct directly from a List[Dim]
        dom2 = tiledb.Domain(dims)
        self.assertEqual(dom, dom2)

        try:
            assert xml.etree.ElementTree.fromstring(dom._repr_html_()) is not None
        except:
            pytest.fail(f"Could not parse dom._repr_html_(). Saw {dom._repr_html_()}")

    def test_datetime_domain(self):
        dim = tiledb.Dim(
            name="d1",
            domain=(np.datetime64("2010-01-01"), np.datetime64("2020-01-01")),
            tile=np.timedelta64(20, "D"),
            dtype=np.datetime64("", "D"),
        )
        dom = tiledb.Domain(dim)
        self.assertEqual(dom, dom)
        self.assertEqual(dom.dtype, np.datetime64("", "D"))

    def test_domain_mixed_names_error(self):
        with self.assertRaises(tiledb.TileDBError):
            tiledb.Domain(
                tiledb.Dim("d1", (1, 4), 2, dtype="u8"),
                tiledb.Dim("__dim_0", (1, 4), 2, dtype="u8"),
            )

    def test_ascii_domain(self, capfd):
        path = self.path("test_ascii_domain")

        dim = tiledb.Dim(name="d", dtype="ascii")
        assert dim.dtype == np.bytes_

        dom = tiledb.Domain(dim)
        self.assertEqual(dom, dom)
        dom.dump()
        assert_captured(capfd, "Type: STRING_ASCII")

        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=True)
        tiledb.SparseArray.create(path, schema)

        ascii_coords = ["a", "b", "c", "ABC"]
        unicode_coords = ["±", "×", "÷", "√"]
        data = [1, 2, 3, 4]

        with tiledb.open(path, "w") as A:
            with self.assertRaises(tiledb.TileDBError):
                A[unicode_coords] = data
            A[ascii_coords] = data



import tempfile
import unittest

import numpy as np
import pytest

import tiledb
import tiledb.libtiledb as lt

if not (lt.version()[0] == 2 and lt.version()[1] >= 25):
    pytest.skip(
        "CurrentDomain is only available in TileDB 2.26 and later",
        allow_module_level=True,
    )


class NDRectangleTest(unittest.TestCase):
    def test_ndrectagle_standalone_string(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", dtype="S0"),
            tiledb.Dim(name="d2", dtype="S0"),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = ("a", "c")
        range_two = ("b", "db")

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        # should be the same
        self.assertEqual(ndrect.range("d1"), range_one)
        self.assertEqual(ndrect.range("d2"), range_two)

    def test_ndrectagle_standalone_integer(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="x", domain=(0, 100), tile=10, dtype=np.int64),
            tiledb.Dim(name="y", domain=(0, 100), tile=10, dtype=np.int64),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = (10, 20)
        range_two = (30, 40)

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        # should be the same
        self.assertEqual(ndrect.range("x"), range_one)
        self.assertEqual(ndrect.range("y"), range_two)


class CurrentDomainTest(unittest.TestCase):
    def test_current_domain_with_ndrectangle_integer(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="x", domain=(0, 100), tile=10, dtype=np.int64),
            tiledb.Dim(name="y", domain=(0, 100), tile=10, dtype=np.int64),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = (10, 20)
        range_two = (30, 40)

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        current_domain = tiledb.CurrentDomain(ctx)

        self.assertTrue(current_domain.is_empty)
        current_domain.set_ndrectangle(ndrect)
        self.assertFalse(current_domain.is_empty)

        # let's try to get the NDRectangle back from the current domain object
        rect = current_domain.ndrectangle

        range1 = rect.range(0)
        range2 = rect.range(1)

        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

        range1 = rect.range("x")
        range2 = rect.range("y")

        # should be the same
        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

    def test_current_domain_with_ndrectangle_string(self):
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d1", dtype="S0"),
            tiledb.Dim(name="d2", dtype="S0"),
        )
        ndrect = tiledb.NDRectangle(ctx, dom)

        range_one = ("a", "c")
        range_two = ("b", "db")

        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        self.assertEqual(ndrect.range(0), range_one)
        self.assertEqual(ndrect.range(1), range_two)

        current_domain = tiledb.CurrentDomain(ctx)

        self.assertTrue(current_domain.is_empty)
        current_domain.set_ndrectangle(ndrect)
        self.assertFalse(current_domain.is_empty)

        # let's try to get the NDRectangle back from the current domain object
        rect = current_domain.ndrectangle

        range1 = rect.range(0)
        range2 = rect.range(1)

        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

        range1 = rect.range("d1")
        range2 = rect.range("d2")

        # should be the same
        self.assertEqual(range1, range_one)
        self.assertEqual(range2, range_two)

    def test_array_schema_with_current_domain_with_ndrectangle(self):
        uri = tempfile.mkdtemp()
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d", domain=(1, 999), tile=2, dtype=np.int64),
            tiledb.Dim(name="d2", domain=(1, 999), tile=2, dtype=np.int64),
        )
        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=True, ctx=ctx, domain=dom, attrs=(att,))

        ndrect = tiledb.NDRectangle(ctx, dom)
        range_one = (10, 20)
        range_two = (30, 40)
        ndrect.set_range(0, range_one[0], range_one[1])
        ndrect.set_range(1, range_two[0], range_two[1])

        current_domain = tiledb.CurrentDomain(ctx)
        current_domain.set_ndrectangle(ndrect)
        schema.set_current_domain(current_domain)

        # create the array
        tiledb.Array.create(uri, schema)

        # open the array and check the current domain and the NDRectangle
        A = tiledb.Array(uri, mode="r")

        cd = A.schema.current_domain
        self.assertFalse(cd.is_empty)
        self.assertEqual(cd.type, lt.CurrentDomainType.NDRECTANGLE)

        ndr = cd.ndrectangle
        self.assertEqual(ndr.range(0), range_one)
        self.assertEqual(ndr.range(1), range_two)

        # a 3rd dimension should raise an error
        with self.assertRaises(tiledb.TileDBError):
            ndr.range(2)

        A.close()

    def test_current_domain_evolve(self):
        uri = tempfile.mkdtemp()
        ctx = tiledb.Ctx()
        dom = tiledb.Domain(
            tiledb.Dim(name="d", domain=(1, 999), tile=2, dtype=np.int64),
        )

        att = tiledb.Attr(name="a", dtype=np.int64)
        schema = tiledb.ArraySchema(sparse=True, ctx=ctx, domain=dom, attrs=(att,))

        ndrect = tiledb.NDRectangle(ctx, dom)
        range_one = (10, 20)
        ndrect.set_range(0, range_one[0], range_one[1])

        current_domain = tiledb.CurrentDomain(ctx)
        current_domain.set_ndrectangle(ndrect)
        schema.set_current_domain(current_domain)

        tiledb.Array.create(uri, schema)

        new_range = (5, 30)
        new_ndrect = tiledb.NDRectangle(ctx, dom)
        new_ndrect.set_range(0, new_range[0], new_range[1])
        new_current_domain = tiledb.CurrentDomain(ctx)
        new_current_domain.set_ndrectangle(new_ndrect)

        se = tiledb.ArraySchemaEvolution(ctx)
        se.expand_current_domain(new_current_domain)
        se.array_evolve(uri)

        A = tiledb.Array(uri, mode="r")
        s = A.schema
        cd = s.current_domain
        n = cd.ndrectangle
        self.assertEqual(n.range(0), new_range)
        A.close()



# writing_dense_rgb.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2021 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#   https://docs.tiledb.com/main/how-to/arrays/writing-arrays/writing-in-dense-subarrays
#
# When run, this program will create a 2D+1 multi-component (eg RGB) dense array, write some
# data to it, and read the entire array data.

import numpy as np

import tiledb

img_shape = (100, 224, 224)
img_uri = "writing_dense_rgb"

image_data = np.random.randint(low=0, high=100, size=(*img_shape, 3), dtype=np.int32)


def create_array():
    domain = tiledb.Domain(
        tiledb.Dim(
            name="image_id", domain=(0, img_shape[0] - 1), tile=4, dtype=np.int32
        ),
        tiledb.Dim(
            name="x", domain=(0, img_shape[1] - 1), tile=img_shape[1], dtype=np.int32
        ),
        tiledb.Dim(
            name="y", domain=(0, img_shape[2] - 1), tile=img_shape[2], dtype=np.int32
        ),
    )

    # create multi-component attribute with three int32 components
    attr = tiledb.Attr(dtype=np.dtype("i4, i4, i4"))

    schema = tiledb.ArraySchema(domain=domain, sparse=False, attrs=[attr])

    tiledb.Array.create(img_uri, schema, overwrite=True)

    image_data_rgb = image_data.view(np.dtype("i4, i4, i4")).reshape(img_shape)

    with tiledb.open(img_uri, "w") as A:
        # write data to 1st image_id slot
        A[:] = image_data_rgb


def read_array():
    with tiledb.open(img_uri) as A:
        print(A[:].shape)


if __name__ == "__main__":
    create_array()
    read_array()



import datetime
import os
import subprocess
import sys

import numpy as np
import pytest

import tiledb
from tiledb.main import PyFragmentInfo
from tiledb.tests.common import DiskTestCase


def has_libfaketime():
    try:
        subprocess.check_output(["which", "faketime"])
        return True
    except subprocess.CalledProcessError:
        return False


@pytest.mark.skipif(
    sys.platform == "win32" or not has_libfaketime(),
    reason=f"libfaketime not installed. {'Not supported on Windows.' if sys.platform == 'win32' else ''}",
)
class TestTimestampOverrides(DiskTestCase):
    def test_timestamp_overrides(self):
        uri_fragments = self.path("time_test_fragments")
        uri_group_metadata = self.path("time_test_group_metadata")

        python_exe = sys.executable
        cmd = (
            f"from tiledb.tests.test_timestamp_overrides import TestTimestampOverrides; "
            f"TestTimestampOverrides().helper_fragments('{uri_fragments}'); "
            f"TestTimestampOverrides().helper_group_metadata('{uri_group_metadata}')"
        )
        test_path = os.path.dirname(os.path.abspath(__file__))

        try:
            # "+x0" is the time multiplier, which makes the time freeze during the test
            subprocess.check_output(
                ["faketime", "-f", "+x0", python_exe, "-c", cmd], cwd=test_path
            )
        except subprocess.CalledProcessError as e:
            raise e

    def helper_fragments(self, uri):
        start_datetime = datetime.datetime.now()

        fragments = 5
        A = np.zeros(fragments)

        dom = tiledb.Domain(tiledb.Dim(domain=(0, 4), tile=fragments, dtype=np.int64))
        att = tiledb.Attr(dtype=A.dtype)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,))

        tiledb.DenseArray.create(uri, schema)

        uris_seen = set()
        chronological_order = []

        for fragment_idx in range(fragments):
            with tiledb.DenseArray(uri, mode="w") as T:
                T[fragment_idx : fragment_idx + 1] = fragment_idx

            # Read the data back immediately after writing to ensure it is correct
            with tiledb.DenseArray(uri, mode="r") as T:
                read_data = T[fragment_idx : fragment_idx + 1]
            self.assertEqual(read_data, np.array([fragment_idx]))

            fragment_info = PyFragmentInfo(uri, schema, False, tiledb.default_ctx())
            uris = fragment_info.get_uri()
            new_uri = set(uris) - uris_seen
            uris_seen.update(uris)
            chronological_order.extend(new_uri)

        end_datetime = datetime.datetime.now()
        self.assertEqual(start_datetime, end_datetime)

        fragment_info = PyFragmentInfo(uri, schema, False, tiledb.default_ctx())
        final_uris = fragment_info.get_uri()

        # Keep only the last part of the uris
        final_uris = [os.path.basename(uri) for uri in final_uris]
        chronological_order = [os.path.basename(uri) for uri in chronological_order]

        # Check that timestamps are the same (faketime is working)
        timestamps = set()
        for uri in final_uris:
            parts = uri.split("_")
            timestamps.add((parts[2], parts[3]))

        self.assertEqual(len(timestamps), 1)

        # Check that UUIDs are unique
        uuids = set()
        for uri in final_uris:
            parts = uri.split("_")
            uuids.add(parts[4])

        self.assertEqual(len(uuids), fragments)

        # Ensure that write order is correct
        self.assertEqual(chronological_order, sorted(final_uris))

    def helper_group_metadata(self, uri):
        vfs = tiledb.VFS()

        start_datetime = datetime.datetime.now()

        tiledb.Group.create(uri)
        loop_count = 10
        uris_seen = set()
        chronological_order = []
        meta_path = f"{uri}/__meta"

        for i in range(loop_count):
            with tiledb.Group(uri, "w") as grp:
                grp.meta["meta"] = i

            # Read the data back immediately after writing to ensure it is correct
            with tiledb.Group(uri, "r") as grp:
                self.assertEqual(grp.meta["meta"], i)

            uris = vfs.ls(meta_path)
            new_uri = set(uris) - uris_seen
            uris_seen.update(uris)
            chronological_order.extend(new_uri)

        end_datetime = datetime.datetime.now()
        self.assertEqual(start_datetime, end_datetime)

        final_uris = vfs.ls(meta_path)

        # Keep only the last part of the uris
        final_uris = [os.path.basename(uri) for uri in final_uris]
        chronological_order = [os.path.basename(uri) for uri in chronological_order]

        # Check that timestamps are the same (faketime is working)
        timestamps = set()
        for uri in final_uris:
            parts = uri.split("_")
            timestamps.add((parts[2], parts[3]))

        self.assertEqual(len(timestamps), 1)

        # Check that UUIDs are unique
        uuids = set()
        for uri in final_uris:
            parts = uri.split("_")
            uuids.add(parts[4])

        self.assertEqual(len(uuids), loop_count)

        # Ensure that write order is correct
        self.assertEqual(chronological_order, sorted(final_uris))



# quickstart_dimension_labels.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2023 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please refer to the TileDB and TileDB-Py documentation for more information:
#   https://docs.tiledb.com/main/how-to
#   https://tiledb-inc-tiledb.readthedocs-hosted.com/projects/tiledb-py/en/stable/python-api.html
#
# When run, this program will create a simple 2D dense array with dimension labels, write
# some data to it, and read a slice of the data back.
#


import numpy as np

import tiledb


def create_array(uri: str):
    """Create array schema with dimension labels"""
    dim1 = tiledb.Dim("d1", domain=(1, 5))
    dim2 = tiledb.Dim("d2", domain=(1, 5))
    dom = tiledb.Domain(dim1, dim2)
    att = tiledb.Attr("a1", dtype=np.int64)
    dim_labels = {
        0: {"l1": dim1.create_label_schema("decreasing", np.int64)},
        1: {
            "l2": dim2.create_label_schema("increasing", np.int64),
            "l3": dim2.create_label_schema("increasing", np.float64),
        },
    }
    schema = tiledb.ArraySchema(domain=dom, attrs=(att,), dim_labels=dim_labels)
    tiledb.Array.create(uri, schema)


def write_array(uri: str):
    """Write attribute and label data to the array"""
    a1_data = np.reshape(np.arange(1, 26), (5, 5))
    l1_data = np.arange(5, 0, -1)
    l2_data = np.arange(-2, 3)
    l3_data = np.linspace(-1.0, 1.0, 5)
    with tiledb.open(uri, "w") as array:
        array[:] = {"a1": a1_data, "l1": l1_data, "l2": l2_data, "l3": l3_data}


def read_array(uri: str):
    """Read the array from the dimension label"""

    with tiledb.open(uri, "r") as array:
        data1 = array.label_index(["l2"])[1, -1:1]
        print("Reading array on [[1, -1:1]] with label 'l2' on dim2")
        for name, value in data1.items():
            print(f"  '{name}'={value}")

        data2 = array.label_index(["l1", "l2"])[4:5, -2:2]
        print("Reading array on [[4:5, -2:2]] with label 'l1' on dim1 and 'l2' on dim2")
        for name, value in data2.items():
            print(f"  '{name}'={value}")

        data3 = array.label_index(["l3"])[1, 0.0:2.0]
        print("Reading array on [[1, 0.0:2.0]] with label 'l3' on dim2")
        for name, value in data3.items():
            print(f"  '{name}'={value}")


if __name__ == "__main__":
    # Name of the array to create.
    ARRAY_NAME = "quickstart_labels"

    LIBVERSION = tiledb.libtiledb.version()

    if LIBVERSION[0] == 2 and LIBVERSION[1] < 15:
        print(
            f"Dimension labels requires libtiledb version >= 2.15.0. Current version is"
            f" {LIBVERSION[0]}.{LIBVERSION[1]}.{LIBVERSION[2]}"
        )

    else:
        # Only create and write to the array if it doesn't already exist.
        if tiledb.object_type(ARRAY_NAME) != "array":
            create_array(ARRAY_NAME)
            write_array(ARRAY_NAME)

        # Read from the array and print output.
        read_array(ARRAY_NAME)



import dataclasses
from typing import Tuple, Union

from . import _generated_version


@dataclasses.dataclass(frozen=True)
class VersionHelper:
    version: str
    version_tuple: Tuple[Union[str, int], ...]

    def __call__(self) -> Tuple[Union[str, int], ...]:
        return self.version_tuple


version = VersionHelper(_generated_version.version, _generated_version.version_tuple)



import warnings
from typing import Dict, List

import numpy as np

import tiledb
import tiledb.libtiledb as lt

from .ctx import Config, Ctx, default_ctx
from .datatypes import DataType
from .domain_indexer import DomainIndexer
from .enumeration import Enumeration
from .metadata import Metadata

# Integer types supported by Python / System
_inttypes = (int, np.integer)


def _tiledb_datetime_extent(begin, end):
    """
    Returns the integer extent of a datetime range.

    :param begin: beginning of datetime range
    :type begin: numpy.datetime64
    :param end: end of datetime range
    :type end: numpy.datetime64
    :return: Extent of range, returned as an integer number of time units
    :rtype: int
    """
    extent = end - begin + 1
    date_unit = np.datetime_data(extent.dtype)[0]
    one = np.timedelta64(1, date_unit)
    # Dividing a timedelta by 1 will convert the timedelta to an integer
    return int(extent / one)


def index_as_tuple(idx):
    """Forces scalar index objects to a tuple representation"""
    if isinstance(idx, tuple):
        return idx
    return (idx,)


def replace_ellipsis(ndim: int, idx: tuple):
    """
    Replace indexing ellipsis object with slice objects to match the number
    of dimensions.
    """
    # count number of ellipsis
    n_ellip = sum(1 for i in idx if i is Ellipsis)
    if n_ellip > 1:
        raise IndexError("an index can only have a single ellipsis ('...')")
    elif n_ellip == 1:
        n = len(idx)
        if (n - 1) >= ndim:
            # does nothing, strip it out
            idx = tuple(i for i in idx if i is not Ellipsis)
        else:
            # locate where the ellipse is, count the number of items to left and right
            # fill in whole dim slices up to th ndim of the array
            left = idx.index(Ellipsis)
            right = n - (left + 1)
            new_idx = idx[:left] + ((slice(None),) * (ndim - (n - 1)))
            if right:
                new_idx += idx[-right:]
            idx = new_idx
    idx_ndim = len(idx)
    if idx_ndim < ndim:
        idx += (slice(None),) * (ndim - idx_ndim)
    if len(idx) > ndim:
        raise IndexError("too many indices for array")
    return idx


def replace_scalars_slice(dom, idx: tuple):
    """Replace scalar indices with slice objects"""
    new_idx, drop_axes = [], []
    for i in range(dom.ndim):
        dim = dom.dim(i)
        dim_idx = idx[i]
        if np.isscalar(dim_idx):
            drop_axes.append(i)
            if isinstance(dim_idx, _inttypes):
                start = int(dim_idx)
                if start < 0:
                    start += int(dim.domain[1]) + 1
                stop = start + 1
            else:
                start = dim_idx
                stop = dim_idx
            new_idx.append(slice(start, stop, None))
        else:
            new_idx.append(dim_idx)
    return tuple(new_idx), tuple(drop_axes)


def check_for_floats(selection):
    """
    Check if a selection object contains floating point values

    :param selection: selection object
    :return: True if selection contains floating point values
    :rtype: bool
    """
    if isinstance(selection, float):
        return True
    if isinstance(selection, slice):
        if isinstance(selection.start, float) or isinstance(selection.stop, float):
            return True
    elif isinstance(selection, tuple):
        for s in selection:
            if check_for_floats(s):
                return True
    return False


def index_domain_subarray(array, dom, idx: tuple):
    """
    Return a numpy array representation of the tiledb subarray buffer
    for a given domain and tuple of index slices
    """
    ndim = dom.ndim
    if len(idx) != ndim:
        raise IndexError(
            "number of indices does not match domain rank: "
            "(got {!r}, expected: {!r})".format(len(idx), ndim)
        )

    subarray = list()
    for r in range(ndim):
        # extract lower and upper bounds for domain dimension extent
        dim = dom.dim(r)
        dim_dtype = dim.dtype

        if array.mode == "r" and (
            np.issubdtype(dim_dtype, np.str_) or np.issubdtype(dim_dtype, np.bytes_)
        ):
            # NED can only be retrieved in read mode
            ned = array.nonempty_domain()
            (dim_lb, dim_ub) = ned[r] if ned else (None, None)
        else:
            (dim_lb, dim_ub) = dim.domain

        dim_slice = idx[r]
        if not isinstance(dim_slice, slice):
            raise IndexError("invalid index type: {!r}".format(type(dim_slice)))

        start, stop, step = dim_slice.start, dim_slice.stop, dim_slice.step

        if np.issubdtype(dim_dtype, np.str_) or np.issubdtype(dim_dtype, np.bytes_):
            if start is None or stop is None:
                if start is None:
                    start = dim_lb
                if stop is None:
                    stop = dim_ub
            elif not isinstance(start, (str, bytes)) or not isinstance(
                stop, (str, bytes)
            ):
                raise tiledb.TileDBError(
                    f"Non-string range '({start},{stop})' provided for string dimension '{dim.name}'"
                )
            subarray.append((start, stop))
            continue

        if step and array.schema.sparse:
            raise IndexError("steps are not supported for sparse arrays")

        # Datetimes will be treated specially
        is_datetime = dim_dtype.kind == "M"

        # Promote to a common type
        if start is not None and stop is not None:
            if type(start) != type(stop):
                promoted_dtype = np.promote_types(type(start), type(stop))
                start = np.array(start, dtype=promoted_dtype, ndmin=1)[0]
                stop = np.array(stop, dtype=promoted_dtype, ndmin=1)[0]

        if start is not None:
            if is_datetime and not isinstance(start, np.datetime64):
                raise IndexError(
                    "cannot index datetime dimension with non-datetime interval"
                )
            # don't round / promote fp slices
            if np.issubdtype(dim_dtype, np.integer):
                if isinstance(start, (np.float32, np.float64)):
                    raise IndexError(
                        "cannot index integral domain dimension with floating point slice"
                    )
                elif not isinstance(start, _inttypes):
                    raise IndexError(
                        "cannot index integral domain dimension with non-integral slice (dtype: {})".format(
                            type(start)
                        )
                    )
            # apply negative indexing (wrap-around semantics)
            if not is_datetime and start < 0:
                start += int(dim_ub) + 1
            if start < dim_lb:
                # numpy allows start value < the array dimension shape,
                # clamp to lower bound of dimension domain
                # start = dim_lb
                raise IndexError("index out of bounds <todo>")
        else:
            start = dim_lb
        if stop is not None:
            if is_datetime and not isinstance(stop, np.datetime64):
                raise IndexError(
                    "cannot index datetime dimension with non-datetime interval"
                )
            # don't round / promote fp slices
            if np.issubdtype(dim_dtype, np.integer):
                if isinstance(start, (np.float32, np.float64)):
                    raise IndexError(
                        "cannot index integral domain dimension with floating point slice"
                    )
                elif not isinstance(start, _inttypes):
                    raise IndexError(
                        "cannot index integral domain dimension with non-integral slice (dtype: {})".format(
                            type(start)
                        )
                    )
            if not is_datetime and stop < 0:
                stop = np.int64(stop) + dim_ub
            if stop > dim_ub:
                # numpy allows stop value > than the array dimension shape,
                # clamp to upper bound of dimension domain
                if is_datetime:
                    stop = dim_ub
                else:
                    stop = int(dim_ub) + 1
        else:
            if np.issubdtype(dim_dtype, np.floating) or is_datetime:
                stop = dim_ub
            else:
                stop = int(dim_ub) + 1

        if np.issubdtype(type(stop), np.floating):
            # inclusive bounds for floating point / datetime ranges
            start = dim_dtype.type(start)
            stop = dim_dtype.type(stop)
            subarray.append((start, stop))
        elif is_datetime:
            # need to ensure that datetime ranges are in the units of dim_dtype
            # so that add_range and output shapes work correctly
            start = start.astype(dim_dtype)
            stop = stop.astype(dim_dtype)
            subarray.append((start, stop))
        elif np.issubdtype(type(stop), np.integer):
            # normal python indexing semantics
            subarray.append((start, int(stop) - 1))
        else:
            raise IndexError(
                "domain indexing is defined for integral and floating point values"
            )
    return subarray


# this function loads the pybind Array to determine whether it is a sparse or dense array.
def preload_array(uri, mode, key, timestamp, ctx=None):
    if ctx is None:
        if key is not None:
            config = tiledb.Config()
            config["sm.encryption_key"] = key
            config["sm.encryption_type"] = "AES_256_GCM"
            ctx = tiledb.Ctx(config=config)
        else:
            ctx = default_ctx()

    _mode_to_query_type = {
        "r": lt.QueryType.READ,
        "w": lt.QueryType.WRITE,
        "m": lt.QueryType.MODIFY_EXCLUSIVE,
        "d": lt.QueryType.DELETE,
    }
    try:
        query_type = _mode_to_query_type[mode]
    except KeyError:
        raise ValueError(
            f"TileDB array mode must be one of {_mode_to_query_type.keys()}"
        )

    ts_start = None
    ts_end = None
    if timestamp is not None:
        if isinstance(timestamp, tuple):
            if len(timestamp) != 2 and not (
                isinstance(timestamp[0], int) and isinstance(timestamp[1], int)
            ):
                raise ValueError(
                    "'timestamp' argument expects either int or tuple(start: int, end: int)"
                )
            ts_start, ts_end = timestamp
        elif isinstance(timestamp, int):
            # handle the existing behavior for unary timestamp
            # which is equivalent to endpoint of the range
            ts_end = timestamp
        else:
            raise TypeError("Unexpected argument type for 'timestamp' keyword argument")

    return lt.Array(ctx, uri, query_type, (ts_start, ts_end))


class Array:
    """Base class for TileDB array objects.

    Defines common properties/functionality for the different array types. When
    an Array instance is initialized, the array is opened with the specified mode.

    :param str uri: URI of array to open
    :param str mode: (default 'r') Open the array object in read 'r', write 'w', modify exclusive 'm', or delete 'd' mode
    :param str key: (default None) If not None, encryption key to decrypt the array
    :param tuple timestamp: (default None) If int, open the array at a given TileDB
        timestamp. If tuple, open at the given start and end TileDB timestamps.
    :param str attr: (default None) open one attribute of the array; indexing a
        dense array will return a Numpy ndarray directly rather than a dictionary.
    :param Ctx ctx: TileDB context
    """

    def __init__(
        self, uri, mode="r", key=None, timestamp=None, attr=None, ctx=None, **kwargs
    ):
        if ctx is None:
            ctx = default_ctx()

        self.array = kwargs.get(
            "preloaded_array", preload_array(uri, mode, key, timestamp, ctx)
        )

        # view on a single attribute
        schema = self.array._schema()
        if attr is not None and not schema._has_attribute(attr):
            self.array._close()
            raise KeyError(f"No attribute matching '{attr}'")
        else:
            self.view_attr = attr

        self.ctx = ctx
        self.key = key
        self.uri = uri
        self.domain_index = DomainIndexer(self)
        self.pyquery = None
        self.__buffers = None

        self.last_fragment_info = dict()
        self._meta = Metadata(self.array)

    def __capsule__(self):
        return self.array.__capsule__()

    def __repr__(self):
        if self.isopen:
            array_type = "Sparse" if self.schema.sparse else "Dense"
            return f"Array(type={array_type }, uri={self.uri!r}, mode={self.mode}, ndim={self.schema.ndim})"
        else:
            return f"Array(uri={self.uri!r}, mode=closed)"

    @classmethod
    def create(cls, uri, schema, key=None, overwrite=False, ctx=None):
        """Creates a TileDB Array at the given URI

        :param str uri: URI at which to create the new empty array.
        :param ArraySchema schema: Schema for the array
        :param str key: (default None) Encryption key to use for array
        :param bool overwrite: (default False) Overwrite the array if it already exists
        :param Ctx ctx: (default None) Optional TileDB Ctx used when creating the array,
                        by default uses the ArraySchema's associated context
                        (*not* necessarily ``tiledb.default_ctx``).

        """
        if ctx is not None and not isinstance(ctx, Ctx):
            raise TypeError(
                "tiledb.Array.create() expected tiledb.Ctx " "object to argument ctx"
            )
        ctx = ctx or default_ctx()

        from .dense_array import DenseArrayImpl
        from .sparse_array import SparseArrayImpl

        if issubclass(cls, DenseArrayImpl) and schema.sparse:
            raise ValueError(
                "Array.create `schema` argument must be a dense schema for DenseArray and subclasses"
            )
        if issubclass(cls, SparseArrayImpl) and not schema.sparse:
            raise ValueError(
                "Array.create `schema` argument must be a sparse schema for SparseArray and subclasses"
            )

        config = tiledb.Config(ctx.config())
        if key is not None:
            config["sm.encryption_type"] = "AES_256_GCM"
            config["sm.encryption_key"] = key
            ctx = tiledb.Ctx(config)

        if overwrite:
            from .highlevel import object_type
            from .vfs import VFS

            if object_type(uri) == "array":
                if uri.startswith("file://") or "://" not in uri:
                    try:
                        VFS(config=config, ctx=ctx).remove_dir(uri)
                    except tiledb.TileDBError as e:
                        raise tiledb.TileDBError(
                            f"Error removing existing array at '{uri}': {e}"
                        )
                else:
                    raise TypeError("Cannot overwrite non-local array.")
            else:
                warnings.warn("Overwrite set, but array does not exist")

        lt.Array._create(ctx, uri, schema)

    @classmethod
    def load_typed(cls, uri, mode="r", key=None, timestamp=None, attr=None, ctx=None):
        """Return a {Dense,Sparse}Array instance from a pre-opened Array (internal)"""
        if ctx is None:
            ctx = default_ctx()

        tmp_array = preload_array(uri, mode, key, timestamp, ctx)

        if tmp_array._schema()._array_type == lt.ArrayType.SPARSE:
            return tiledb.SparseArray(
                uri, mode, key, timestamp, attr, ctx, preloaded_array=tmp_array
            )
        else:
            return tiledb.DenseArray(
                uri, mode, key, timestamp, attr, ctx, preloaded_array=tmp_array
            )

    def __enter__(self):
        """
        The `__enter__` and `__exit__` methods allow TileDB arrays to be opened (and auto-closed)
        using `with tiledb.open(uri) as A:` syntax.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        The `__enter__` and `__exit__` methods allow TileDB arrays to be opened (and auto-closed)
        using `with tiledb.open(uri) as A:` syntax.
        """
        self.close()

    def close(self):
        """Closes this array, flushing all buffered data."""
        self.array._close()

    def reopen(self):
        """
        Reopens this array.

        This is useful when the array is updated after it was opened.
        To sync-up with the updates, the user must either close the array and open again,
        or just use ``reopen()`` without closing. ``reopen`` will be generally faster than
        a close-then-open.
        """
        self.array._reopen()

    @property
    def schema(self):
        """The :py:class:`ArraySchema` for this array."""
        schema = self.array._schema()
        if schema is None:
            raise tiledb.TileDBError("Cannot access schema, array is closed")
        from .array_schema import ArraySchema

        return ArraySchema.from_pybind11(self.ctx, schema)

    @property
    def mode(self):
        """The mode this array was opened with."""
        mode = self.array._query_type()
        _query_type_to_mode = {
            lt.QueryType.READ: "r",
            lt.QueryType.WRITE: "w",
            lt.QueryType.MODIFY_EXCLUSIVE: "m",
            lt.QueryType.DELETE: "d",
        }
        return _query_type_to_mode[mode]

    @property
    def iswritable(self):
        """This array is currently opened as writable."""
        return self.mode == "w"

    @property
    def isopen(self):
        """True if this array is currently open."""
        return self.array._is_open()

    @property
    def ndim(self):
        """The number of dimensions of this array."""
        return self.schema.ndim

    @property
    def domain(self):
        """The :py:class:`Domain` of this array."""
        return self.schema.domain

    @property
    def dtype(self):
        """The NumPy dtype of the specified attribute"""
        if self.view_attr is None and self.schema.nattr > 1:
            raise NotImplementedError("Multi-attribute does not have single dtype!")
        return self.schema.attr(0).dtype

    @property
    def shape(self):
        """The shape of this array."""
        return self.schema.shape

    @property
    def nattr(self):
        """The number of attributes of this array."""
        return 1 if self.view_attr is not None else self.schema.nattr

    def view_attr(self):
        """The view attribute of this array."""
        return self.view_attr

    @property
    def timestamp_range(self):
        """Returns the timestamp range the array is opened at

        :rtype: tuple
        :returns: tiledb timestamp range at which point the array was opened

        """
        timestamp_start = self.array._open_timestamp_start
        timestamp_end = self.array._open_timestamp_end

        return (timestamp_start, timestamp_end)

    @property
    def meta(self) -> Metadata:
        """
        :return: The Array's metadata as a key-value structure
        :rtype: Metadata
        """
        return self._meta

    def subarray(self, selection, attrs=None, coords=False, order=None):
        raise NotImplementedError()

    def attr(self, key):
        """Returns an :py:class:`Attr` instance given an int index or string label

        :param key: attribute index (positional or associative)
        :type key: int or str
        :rtype: :py:class:`Attr`
        :return: The array attribute at index or with the given name (label)
        :raises TypeError: invalid key type"""
        return self.schema.attr(key)

    def dim(self, dim_id):
        """Returns a :py:class:`Dim` instance given a dim index or name

        :param key: attribute index (positional or associative)
        :type key: int or str
        :rtype: :py:class:`Attr`
        :return: The array attribute at index or with the given name (label)
        :raises TypeError: invalid key type"""
        return self.schema.domain.dim(dim_id)

    def enum(self, name):
        """
        Return the Enumeration from the attribute name.

        :param name: attribute name
        :type key: str
        :rtype: `Enumeration`
        """
        enum = self.array._get_enumeration(self.ctx, name)
        return Enumeration.from_pybind11(self.ctx, enum)

    @staticmethod
    def delete_fragments(uri, timestamp_start, timestamp_end, ctx=None):
        """
        Delete a range of fragments from timestamp_start to timestamp_end.
        The array needs to be opened in 'm' mode as shown in the example below.

        :param timestamp_start: the first fragment to delete in the range
        :type timestamp_start: int
        :param timestamp_end: the last fragment to delete in the range
        :type timestamp_end: int

        **Example:**

        >>> import tiledb, tempfile, numpy as np
        >>> path = tempfile.mkdtemp()

        >>> with tiledb.open(path, 'w', timestamp=2) as A:
        ...     A[:] = np.ones(4, dtype=np.int64)

        >>> with tiledb.open(path, 'r') as A:
        ...     A[:]
        array([1., 1., 1., 1.])

        >>> tiledb.Array.delete_fragments(path, 2, 2)

        >>> with tiledb.open(path, 'r') as A:
        ...     A[:]
        array([0., 0., 0., 0.])

        """
        if isinstance(uri, str):
            lt.Array._delete_fragments(
                ctx or default_ctx(), uri, timestamp_start, timestamp_end
            )
        else:
            raise TypeError("uri must be a string")

    @staticmethod
    def delete_array(uri, ctx=None):
        """
        Delete the given array.

        :param str uri: The URI of the array
        :param Ctx ctx: TileDB context

        **Example:**

        >>> import tiledb, tempfile, numpy as np
        >>> path = tempfile.mkdtemp()

        >>> with tiledb.from_numpy(path, np.zeros(4), timestamp=1) as A:
        ...     pass
        >>> tiledb.array_exists(path)
        True

        >>> tiledb.Array.delete_array(path)

        >>> tiledb.array_exists(path)
        False

        """
        lt.Array._delete_array(ctx or default_ctx(), uri)

    def nonempty_domain(self):
        """Return the minimum bounding domain which encompasses nonempty values.

        :rtype: tuple(tuple(numpy scalar, numpy scalar), ...)
        :return: A list of (inclusive) domain extent tuples, that contain all
            nonempty cells

        """
        results = list()
        dom = self.schema.domain

        for dim_idx in range(dom.ndim):
            dim_dtype = dom.dim(dim_idx).dtype

            if self.array._non_empty_domain_is_empty(dim_idx, dim_dtype, self.ctx):
                results.append((None, None))
                continue

            res_x, res_y = self.array._non_empty_domain(dim_idx, dim_dtype)

            if dim_dtype == np.bytes_:
                # convert to bytes if needed
                res_x = bytes(res_x, encoding="utf8")
                res_y = bytes(res_y, encoding="utf8")

            if np.issubdtype(dim_dtype, np.datetime64):
                # convert to np.datetime64
                date_unit = np.datetime_data(dim_dtype)[0]
                res_x = np.datetime64(res_x, date_unit)
                res_y = np.datetime64(res_y, date_unit)

            results.append((res_x, res_y))

        # if result only contains (None, None) tuples just return None
        for x in results:
            if x != (None, None):
                return tuple(results)

        return None

    def consolidate(self, config=None, key=None, fragment_uris=None, timestamp=None):
        """
        Consolidates fragments of an array object for increased read performance.

        Overview: https://docs.tiledb.com/main/concepts/internal-mechanics/consolidation

        :param tiledb.Config config: The TileDB Config with consolidation parameters set
        :param key: (default None) encryption key to decrypt an encrypted array
        :type key: str or bytes
        :param fragment_uris: (default None) Consolidate the array using a list of fragment _names_ (note: the `__ts1_ts2_<label>_<ver>` fragment name form alone, not the full path(s))
        :param timestamp: (default None) If not None, consolidate the array using the given tuple(int, int) UNIX seconds range (inclusive). This argument will be ignored if `fragment_uris` is passed.
        :type timestamp: tuple (int, int)
        :raises: :py:exc:`tiledb.TileDBError`

        Rather than passing the timestamp into this function, it may be set with
        the config parameters `"sm.vacuum.timestamp_start"`and
        `"sm.vacuum.timestamp_end"` which takes in a time in UNIX seconds. If both
        are set then this function's `timestamp` argument will be used.

        """
        if self.mode == "r":
            raise tiledb.TileDBError(
                "cannot consolidate array opened in readonly mode (mode='r')"
            )

        config = tiledb.Config(config or self.ctx.config())

        if key is not None:
            config["sm.encryption_type"] = "AES_256_GCM"
            config["sm.encryption_key"] = key

        if fragment_uris is not None:
            if timestamp is not None:
                warnings.warn(
                    "The `timestamp` argument will be ignored and only fragments "
                    "passed to `fragment_uris` will be consolidate",
                    DeprecationWarning,
                )
            lt.Array._consolidate(self.array.uri, self.ctx, fragment_uris, config)
            return
        elif timestamp is not None:
            warnings.warn(
                "The `timestamp` argument is deprecated; pass a list of "
                "fragment URIs to consolidate with `fragment_uris`",
                DeprecationWarning,
            )

            if (
                not isinstance(timestamp, tuple)
                and len(timestamp) != 2
                and not (
                    isinstance(timestamp[0], int) and isinstance(timestamp[1], int)
                )
            ):
                raise TypeError(
                    "'timestamp' argument expects tuple(start: int, end: int)"
                )

            if timestamp[0] is not None:
                config["sm.consolidation.timestamp_start"] = timestamp[0]
            if timestamp[1] is not None:
                config["sm.consolidation.timestamp_end"] = timestamp[1]

        lt.Array._consolidate(self.uri, self.ctx, config)

    def upgrade_version(self, config=None):
        """
        Upgrades an array to the latest format version.

        :param config: (default None) Configuration parameters for the upgrade
            (`nullptr` means default, which will use the config from `ctx`).
        :raises: :py:exc:`tiledb.TileDBError`
        """

        lt.Array._upgrade_version(self.ctx, self.uri, config)

    @property
    def ptr(self):
        """Return the underlying C++ TileDB array object pointer"""
        return self.array.ptr

    def dump(self):
        print(self.schema._dump(), "\n")

    def domain_index(self):
        return self.domain_index

    @property
    def dindex(self):
        return self.domain_index

    def _write_array(
        self,
        subarray,
        coordinates: List,
        buffer_names: List,
        values: List,
        labels: Dict,
        nullmaps: Dict,
        issparse: bool,
    ):
        # used for buffer conversion (local import to avoid circularity)
        from .main import array_to_buffer

        isfortran = False
        nattr = len(buffer_names)
        nlabel = len(labels)

        # Create arrays to hold buffer sizes
        nbuffer = nattr + nlabel
        if issparse:
            nbuffer += self.schema.ndim
        buffer_sizes = np.zeros((nbuffer,), dtype=np.uint64)

        # Create lists for data and offset buffers
        output_values = list()
        output_offsets = list()

        # Set data and offset buffers for attributes
        for i in range(nattr):
            attr = self.schema.attr(i)
            # if dtype is ASCII, ensure all characters are valid
            if attr.isascii:
                try:
                    values[i] = np.asarray(values[i], dtype=np.bytes_)
                except Exception as exc:
                    raise tiledb.TileDBError(
                        f'dtype of attr {attr.name} is "ascii" but attr_val contains invalid ASCII characters'
                    )

            if attr.isvar:
                try:
                    if attr.isnullable:
                        if np.issubdtype(attr.dtype, np.str_) or np.issubdtype(
                            attr.dtype, np.bytes_
                        ):
                            attr_val = np.array(
                                ["" if v is None else v for v in values[i]]
                            )
                        else:
                            attr_val = np.nan_to_num(values[i])
                    else:
                        attr_val = values[i]
                    buffer, offsets = array_to_buffer(attr_val, True, False)
                except Exception as exc:
                    raise type(exc)(
                        f"Failed to convert buffer for attribute: '{attr.name}'"
                    ) from exc
            else:
                buffer, offsets = values[i], None

            buffer_sizes[i] = buffer.nbytes // (attr.dtype.itemsize or 1)
            output_values.append(buffer)
            output_offsets.append(offsets)

        # Check value layouts
        if len(values) and nattr > 1:
            value = output_values[0]
            isfortran = value.ndim > 1 and value.flags.f_contiguous
            for value in values:
                if value.ndim > 1 and value.flags.f_contiguous and not isfortran:
                    raise ValueError("mixed C and Fortran array layouts")

        # Set data and offsets buffers for dimensions (sparse arrays only)
        ibuffer = nattr
        if issparse:
            for dim_idx, coords in enumerate(coordinates):
                dim = self.schema.domain.dim(dim_idx)
                if dim.isvar:
                    buffer, offsets = array_to_buffer(coords, True, False)
                else:
                    buffer, offsets = coords, None
                buffer_sizes[ibuffer] = buffer.nbytes // (dim.dtype.itemsize or 1)
                output_values.append(buffer)
                output_offsets.append(offsets)

                name = dim.name
                buffer_names.append(name)

                ibuffer = ibuffer + 1

        for label_name, label_values in labels.items():
            # Append buffer name
            buffer_names.append(label_name)
            # Get label data buffer and offsets buffer for the labels
            dim_label = self.schema.dim_label(label_name)
            if dim_label.isvar:
                buffer, offsets = array_to_buffer(label_values, True, False)
            else:
                buffer, offsets = label_values, None
            buffer_sizes[ibuffer] = buffer.nbytes // (dim_label.dtype.itemsize or 1)
            # Append the buffers
            output_values.append(buffer)
            output_offsets.append(offsets)

            ibuffer = ibuffer + 1

        # Allocate the query
        ctx = lt.Context(self.ctx)
        q = lt.Query(ctx, self.array, lt.QueryType.WRITE)

        # Set the layout
        q.layout = (
            lt.LayoutType.UNORDERED
            if issparse
            else (lt.LayoutType.COL_MAJOR if isfortran else lt.LayoutType.ROW_MAJOR)
        )

        # Create and set the subarray for the query (dense arrays only)
        if not issparse:
            q.set_subarray(subarray)

        # Set buffers on the query
        for i, buffer_name in enumerate(buffer_names):
            # Set data buffer
            ncells = DataType.from_numpy(output_values[i].dtype).ncells
            q.set_data_buffer(
                buffer_name,
                output_values[i],
                np.uint64(buffer_sizes[i] * ncells),
            )

            # Set offsets buffer
            if output_offsets[i] is not None:
                output_offsets[i] = output_offsets[i].astype(np.uint64)
                q.set_offsets_buffer(
                    buffer_name, output_offsets[i], output_offsets[i].size
                )

            # Set validity buffer
            if buffer_name in nullmaps:
                nulmap = nullmaps[buffer_name]
                q.set_validity_buffer(buffer_name, nulmap, nulmap.size)

        q._submit()
        q.finalize()

        fragment_info = self.last_fragment_info
        if fragment_info != False:
            if not isinstance(fragment_info, dict):
                raise ValueError(
                    f"Expected fragment_info to be a dict, got {type(fragment_info)}"
                )
            fragment_info.clear()

            result = dict()
            num_fragments = q.fragment_num()

            if num_fragments < 1:
                return result

            for fragment_idx in range(0, num_fragments):
                fragment_uri = q.fragment_uri(fragment_idx)
                fragment_t1, fragment_t2 = q.fragment_timestamp_range(fragment_idx)
                result[fragment_uri] = (fragment_t1, fragment_t2)

            fragment_info.update(result)

    def label_index(self, labels):
        """Retrieve data cells with multi-range, domain-inclusive indexing by label.
        Returns the cross-product of the ranges.

        Accepts a scalar, ``slice``, or list of scalars per-label for querying on the
        corresponding dimensions. For multidimensional arrays querying by labels only on
        a subset of dimensions, ``:`` should be passed in-place for any labels preceeding
        custom ranges.

        ** Example **

        >>> import tiledb, numpy as np, tempfile
        >>> from collections import OrderedDict
        >>> dim1 = tiledb.Dim("d1", domain=(1, 4))
        >>> dim2 = tiledb.Dim("d2", domain=(1, 3))
        >>> dom = tiledb.Domain(dim1, dim2)
        >>> att = tiledb.Attr("a1", dtype=np.int64)
        >>> dim_labels = {
        ...     0: {"l1": dim1.create_label_schema("decreasing", np.int64)},
        ...     1: {
        ...         "l2": dim2.create_label_schema("increasing", np.int64),
        ...         "l3": dim2.create_label_schema("increasing", np.float64),
        ...     },
        ... }
        >>> schema = tiledb.ArraySchema(domain=dom, attrs=(att,), dim_labels=dim_labels)
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     tiledb.Array.create(tmp, schema)
        ...
        ...     a1_data = np.reshape(np.arange(1, 13), (4, 3))
        ...     l1_data = np.arange(4, 0, -1)
        ...     l2_data = np.arange(-1, 2)
        ...     l3_data = np.linspace(0, 1.0, 3)
        ...
        ...     with tiledb.open(tmp, "w") as A:
        ...         A[:] = {"a1": a1_data, "l1": l1_data, "l2": l2_data, "l3": l3_data}
        ...
        ...     with tiledb.open(tmp, "r") as A:
        ...         np.testing.assert_equal(
        ...             A.label_index(["l1"])[3:4],
        ...             OrderedDict({"l1": [4, 3], "a1": [[1, 2, 3], [4, 5, 6]]}),
        ...         )
        ...         np.testing.assert_equal(
        ...             A.label_index(["l1", "l3"])[2, 0.5:1.0],
        ...             OrderedDict(
        ...                 {"l3": [0.5, 1.0], "l1": [2], "a1": [[8, 9]]}
        ...             ),
        ...         )
        ...         np.testing.assert_equal(
        ...             A.label_index(["l2"])[:, -1:0],
        ...             OrderedDict(
        ...                 {"l2": [-1, 0],
        ...                 "a1": [[1, 2], [4, 5], [7, 8], [10, 11]]},
        ...             ),
        ...         )
        ...         np.testing.assert_equal(
        ...             A.label_index(["l3"])[:, 0.5:1.0],
        ...             OrderedDict(
        ...                 {"l3": [0.5, 1.],
        ...                 "a1": [[2, 3], [5, 6], [8, 9], [11, 12]]},
        ...             ),
        ...         )

        :param labels: List of labels to use when querying. Can only use at most one
            label per dimension.
        :param list selection: Per dimension, a scalar, ``slice``, or  list of scalars.
            Each item is iterpreted as a point (scalar) or range (``slice``) used to
            query the array on the corresponding dimension.
        :returns: dict of {'label/attribute': result}.
        :raises: :py:exc:`tiledb.TileDBError`

        """
        # Delayed to avoid circular import
        from .multirange_indexing import LabelIndexer

        return LabelIndexer(self, tuple(labels))

    @property
    def multi_index(self):
        """Retrieve data cells with multi-range, domain-inclusive indexing. Returns
        the cross-product of the ranges.

        :param list selection: Per dimension, a scalar, ``slice``, or list of scalars
            or ``slice`` objects. Scalars and ``slice`` components should match the
            type of the underlying Dimension.
        :returns: dict of {'attribute': result}. Coords are included by default for
            Sparse arrays only (use `Array.query(coords=<>)` to select).
        :raises IndexError: invalid or unsupported index selection
        :raises: :py:exc:`tiledb.TileDBError`

        ``multi_index[]`` accepts, for each dimension, a scalar, ``slice``, or list
        of scalars or ``slice`` objects. Each item is interpreted as a point
        (scalar) or range (``slice``) used to query the array on the corresponding
        dimension.

        Unlike NumPy array indexing, ``multi_index`` respects TileDB's range semantics:
        slice ranges are *inclusive* of the start- and end-point, and negative ranges
        do not wrap around (because a TileDB dimensions may have a negative domain).

        See also: https://docs.tiledb.com/main/api-usage/reading-arrays/multi-range-subarrays

        ** Example **

        >>> import tiledb, tempfile, numpy as np
        >>>
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...    A = tiledb.from_numpy(tmp, np.eye(4) * [1,2,3,4])
        ...    A.multi_index[1]  # doctest: +ELLIPSIS
        ...    A.multi_index[1,1]  # doctest: +ELLIPSIS
        ...    # return row 0 and 2
        ...    A.multi_index[[0,2]]  # doctest: +ELLIPSIS
        ...    # return rows 0 and 2 intersecting column 2
        ...    A.multi_index[[0,2], 2]  # doctest: +ELLIPSIS
        ...    # return rows 0:2 intersecting columns 0:2
        ...    A.multi_index[slice(0,2), slice(0,2)]  # doctest: +ELLIPSIS
        OrderedDict(...''... array([[0., 2., 0., 0.]])...)
        OrderedDict(...''... array([[2.]])...)
        OrderedDict(...''... array([[1., 0., 0., 0.],
                [0., 0., 3., 0.]])...)
        OrderedDict(...''... array([[0.],
                [3.]])...)
        OrderedDict(...''... array([[1., 0., 0.],
                [0., 2., 0.],
                [0., 0., 3.]])...)

        """
        # Delayed to avoid circular import
        from .multirange_indexing import MultiRangeIndexer

        return MultiRangeIndexer(self)

    @property
    def df(self):
        """Retrieve data cells as a Pandas dataframe, with multi-range,
        domain-inclusive indexing using ``multi_index``.

        :param list selection: Per dimension, a scalar, ``slice``, or list of scalars
            or ``slice`` objects. Scalars and ``slice`` components should match the
            type of the underlying Dimension.
        :returns: dict of {'attribute': result}. Coords are included by default for
            Sparse arrays only (use `Array.query(coords=<>)` to select).
        :raises IndexError: invalid or unsupported index selection
        :raises: :py:exc:`tiledb.TileDBError`

        ``df[]`` accepts, for each dimension, a scalar, ``slice``, or list
        of scalars or ``slice`` objects. Each item is interpreted as a point
        (scalar) or range (``slice``) used to query the array on the corresponding
        dimension.

        ** Example **

        >>> import tiledb, tempfile, numpy as np, pandas as pd
        >>>
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...    data = {'col1_f': np.arange(0.0,1.0,step=0.1), 'col2_int': np.arange(10)}
        ...    df = pd.DataFrame.from_dict(data)
        ...    tiledb.from_pandas(tmp, df)
        ...    A = tiledb.open(tmp)
        ...    A.df[1]
        ...    A.df[1:5]
              col1_f  col2_int
           1     0.1         1
              col1_f  col2_int
           1     0.1         1
           2     0.2         2
           3     0.3         3
           4     0.4         4
           5     0.5         5

        """
        # Delayed to avoid circular import
        from .multirange_indexing import DataFrameIndexer

        return DataFrameIndexer(self, use_arrow=None)

    @property
    def last_write_info(self):
        return self.last_fragment_info

    @property
    def _buffers(self):
        return self.__buffers

    def _set_buffers(self, buffers):
        """
        Helper function to set external buffers in the form of
            {'attr_name': (data_array, offsets_array)}
        Buffers will be used to satisfy the next index/query request.
        """
        self.__buffers = buffers

    def set_query(self, serialized_query):
        from .main import PyQuery

        q = PyQuery(self.ctx, self, ("",), (), 0, False)
        q.set_serialized_query(serialized_query)
        q.submit()

        from collections import OrderedDict

        results = q.results()

        out = OrderedDict()
        for name in results.keys():
            arr = results[name][0]
            arr.dtype = q.buffer_dtype(name)
            out[name] = arr
        return out

    # pickling support: this is a lightweight pickle for distributed use.
    #   simply treat as wrapper around URI, not actual data.
    def __getstate__(self):
        config_dict = self.ctx.config().dict()
        return (
            self.uri,
            self.mode,
            self.key,
            self.view_attr,
            self.timestamp_range,
            config_dict,
        )

    def __setstate__(self, state):
        uri, mode, key, view_attr, timestamp_range, config_dict = state

        if config_dict is not {}:
            config = Config(params=config_dict)
            ctx = Ctx(config)
        else:
            ctx = default_ctx()

        self.__init__(
            uri, mode=mode, key=key, attr=view_attr, timestamp=timestamp_range, ctx=ctx
        )



import numpy as np
import pytest

import tiledb

from .common import DiskTestCase


class AggregateTest(DiskTestCase):
    @pytest.mark.parametrize("sparse", [True, False])
    @pytest.mark.parametrize(
        "dtype",
        [
            np.uint8,
            np.int8,
            np.uint16,
            np.int16,
            np.uint32,
            np.int32,
            np.uint64,
            np.int64,
            np.float32,
            np.float64,
        ],
    )
    def test_basic(self, sparse, dtype):
        path = self.path("test_basic")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [tiledb.Attr(name="a", dtype=dtype)]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=sparse)
        tiledb.Array.create(path, schema)

        data = np.random.randint(1, 10, size=10)

        with tiledb.open(path, "w") as A:
            if sparse:
                A[np.arange(0, 10)] = data
            else:
                A[:] = data

        all_aggregates = ("count", "sum", "min", "max", "mean")

        with tiledb.open(path, "r") as A:
            # entire column
            q = A.query()
            expected = q[:]["a"]

            with pytest.raises(tiledb.TileDBError):
                q.agg("bad")[:]

            with pytest.raises(tiledb.TileDBError):
                q.agg("null_count")[:]

            with pytest.raises(NotImplementedError):
                q.agg("count").df[:]

            assert q.agg("sum")[:] == sum(expected)
            assert q.agg("min")[:] == min(expected)
            assert q.agg("max")[:] == max(expected)
            assert q.agg("mean")[:] == sum(expected) / len(expected)
            assert q.agg("count")[:] == len(expected)

            assert q.agg({"a": "sum"})[:] == sum(expected)
            assert q.agg({"a": "min"})[:] == min(expected)
            assert q.agg({"a": "max"})[:] == max(expected)
            assert q.agg({"a": "mean"})[:] == sum(expected) / len(expected)
            assert q.agg({"a": "count"})[:] == len(expected)

            actual = q.agg(all_aggregates)[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            actual = q.agg({"a": all_aggregates})[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            # subarray
            expected = A[4:7]["a"]

            assert q.agg("sum")[4:7] == sum(expected)
            assert q.agg("min")[4:7] == min(expected)
            assert q.agg("max")[4:7] == max(expected)
            assert q.agg("mean")[4:7] == sum(expected) / len(expected)
            assert q.agg("count")[4:7] == len(expected)

            assert q.agg({"a": "sum"})[4:7] == sum(expected)
            assert q.agg({"a": "min"})[4:7] == min(expected)
            assert q.agg({"a": "max"})[4:7] == max(expected)
            assert q.agg({"a": "mean"})[4:7] == sum(expected) / len(expected)
            assert q.agg({"a": "count"})[4:7] == len(expected)

            actual = q.agg(all_aggregates)[4:7]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            actual = q.agg({"a": all_aggregates})[4:7]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

    @pytest.mark.parametrize("sparse", [True, False])
    @pytest.mark.parametrize(
        "dtype",
        [
            np.uint8,
            np.int8,
            np.uint16,
            np.int16,
            np.uint32,
            np.int32,
            np.uint64,
            np.int64,
            np.float32,
            np.float64,
        ],
    )
    def test_multi_index(self, sparse, dtype):
        path = self.path("test_multi_index")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [tiledb.Attr(name="a", dtype=dtype)]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=sparse)
        tiledb.Array.create(path, schema)

        data = np.random.randint(1, 10, size=10)

        with tiledb.open(path, "w") as A:
            if sparse:
                A[np.arange(0, 10)] = data
            else:
                A[:] = data

        all_aggregates = ("count", "sum", "min", "max", "mean")

        with tiledb.open(path, "r") as A:
            # entire column
            q = A.query()
            expected = q.multi_index[:]["a"]

            with pytest.raises(tiledb.TileDBError):
                q.agg("bad")[:]

            with pytest.raises(tiledb.TileDBError):
                q.agg("null_count")[:]

            assert q.agg("sum").multi_index[:] == sum(expected)
            assert q.agg("min").multi_index[:] == min(expected)
            assert q.agg("max").multi_index[:] == max(expected)
            assert q.agg("mean").multi_index[:] == sum(expected) / len(expected)
            assert q.agg("count").multi_index[:] == len(expected)

            actual = q.agg(all_aggregates).multi_index[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            actual = q.agg({"a": all_aggregates}).multi_index[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            # subarray
            expected = A.multi_index[4:7]["a"]

            assert q.agg("sum").multi_index[4:7] == sum(expected)
            assert q.agg("min").multi_index[4:7] == min(expected)
            assert q.agg("max").multi_index[4:7] == max(expected)
            assert q.agg("mean").multi_index[4:7] == sum(expected) / len(expected)
            assert q.agg("count").multi_index[4:7] == len(expected)

            actual = q.agg(all_aggregates).multi_index[4:7]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

    @pytest.mark.parametrize(
        "dtype",
        [
            np.uint8,
            np.int8,
            np.uint16,
            np.int16,
            np.uint32,
            np.int32,
            np.uint64,
            np.int64,
            np.float32,
            np.float64,
        ],
    )
    def test_with_query_condition(self, dtype):
        path = self.path("test_with_query_condition")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [tiledb.Attr(name="a", dtype=dtype)]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=True)
        tiledb.Array.create(path, schema)

        with tiledb.open(path, "w") as A:
            # hardcode the first value to be 1 to ensure that the a < 5
            # query condition always returns a non-empty result
            data = np.random.randint(1, 10, size=10)
            data[0] = 1

            A[np.arange(0, 10)] = data

        all_aggregates = ("count", "sum", "min", "max", "mean")

        with tiledb.open(path, "r") as A:
            q = A.query(cond="a < 5")

            expected = q[:]["a"]
            actual = q.agg(all_aggregates)[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            expected = q.multi_index[:]["a"]
            actual = q.agg(all_aggregates).multi_index[:]
            assert actual["sum"] == sum(expected)
            assert actual["min"] == min(expected)
            assert actual["max"] == max(expected)
            assert actual["mean"] == sum(expected) / len(expected)
            assert actual["count"] == len(expected)

            # no value matches query condition
            q = A.query(cond="a > 10")

            expected = q[:]
            actual = q.agg(all_aggregates)[:]
            assert actual["sum"] == 0
            if dtype in (np.float32, np.float64):
                assert np.isnan(actual["min"])
                assert np.isnan(actual["max"])
            else:
                assert actual["min"] is None
                assert actual["max"] is None
            assert np.isnan(actual["mean"])
            assert actual["count"] == 0

            expected = q.multi_index[:]
            actual = q.agg(all_aggregates).multi_index[:]
            assert actual["sum"] == 0
            if dtype in (np.float32, np.float64):
                assert np.isnan(actual["min"])
                assert np.isnan(actual["max"])
            else:
                assert actual["min"] is None
                assert actual["max"] is None
            assert np.isnan(actual["mean"])
            assert actual["count"] == 0

    @pytest.mark.parametrize("sparse", [True, False])
    def test_nullable(self, sparse):
        path = self.path("test_nullable")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [
            tiledb.Attr(name="integer", nullable=True, dtype=int),
            tiledb.Attr(name="float", nullable=True, dtype=float),
        ]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=sparse)
        tiledb.Array.create(path, schema)

        # set index 5 and 7 to be null
        data = np.random.rand(10)
        data[5], data[7] = np.nan, np.nan

        # write data
        with tiledb.open(path, "w") as A:
            if sparse:
                A[np.arange(0, 10)] = {"integer": data, "float": data}
            else:
                A[:] = {"integer": data, "float": data}

        with tiledb.open(path, "r") as A:
            agg = A.query().agg

            result = agg("null_count")
            assert result[0]["integer"]["null_count"] == 0
            assert result[:6]["integer"]["null_count"] == 1
            assert result[5:8]["integer"]["null_count"] == 2
            assert result[5]["integer"]["null_count"] == 1
            assert result[6:]["integer"]["null_count"] == 1
            assert result[7]["integer"]["null_count"] == 1
            assert result[:]["integer"]["null_count"] == 2

            assert result[0]["float"]["null_count"] == 0
            assert result[:6]["float"]["null_count"] == 1
            assert result[5:8]["float"]["null_count"] == 2
            assert result[5]["float"]["null_count"] == 1
            assert result[6:]["float"]["null_count"] == 1
            assert result[7]["float"]["null_count"] == 1
            assert result[:]["float"]["null_count"] == 2

            all_aggregates = ("count", "sum", "min", "max", "mean", "null_count")

            actual = agg({"integer": all_aggregates, "float": all_aggregates})[:]

            expected = A[:]["integer"]
            expected_no_null = A[:]["integer"].compressed()
            assert actual["integer"]["sum"] == sum(expected_no_null)
            assert actual["integer"]["min"] == min(expected_no_null)
            assert actual["integer"]["max"] == max(expected_no_null)
            assert actual["integer"]["mean"] == sum(expected_no_null) / len(
                expected_no_null
            )
            assert actual["integer"]["count"] == len(expected)
            assert actual["integer"]["null_count"] == np.count_nonzero(expected.mask)

            expected = A[:]["float"]
            expected_no_null = A[:]["float"].compressed()
            assert actual["float"]["sum"] == sum(expected_no_null)
            assert actual["float"]["min"] == min(expected_no_null)
            assert actual["float"]["max"] == max(expected_no_null)
            assert actual["float"]["mean"] == sum(expected_no_null) / len(
                expected_no_null
            )
            assert actual["float"]["count"] == len(expected)
            assert actual["float"]["null_count"] == np.count_nonzero(expected.mask)

            # no valid values
            actual = agg({"integer": all_aggregates, "float": all_aggregates})[5]

            assert actual["integer"]["sum"] is None
            assert actual["integer"]["min"] is None
            assert actual["integer"]["max"] is None
            assert actual["integer"]["mean"] is None
            assert actual["integer"]["count"] == 1
            assert actual["integer"]["null_count"] == 1

            assert np.isnan(actual["float"]["sum"])
            assert np.isnan(actual["float"]["min"])
            assert np.isnan(actual["float"]["max"])
            assert np.isnan(actual["float"]["mean"])
            assert actual["float"]["count"] == 1
            assert actual["float"]["null_count"] == 1

    @pytest.mark.parametrize("sparse", [True, False])
    def test_empty(self, sparse):
        path = self.path("test_empty_sparse")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [
            tiledb.Attr(name="integer", nullable=True, dtype=int),
            tiledb.Attr(name="float", nullable=True, dtype=float),
        ]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=sparse)
        tiledb.Array.create(path, schema)

        data = np.random.rand(5)

        # write data
        with tiledb.open(path, "w") as A:
            if sparse:
                A[np.arange(0, 5)] = {"integer": data, "float": data}
            else:
                A[:5] = {"integer": data, "float": data}

        with tiledb.open(path, "r") as A:
            invalid_aggregates = ("sum", "min", "max", "mean")
            actual = A.query().agg(invalid_aggregates)[6:]

            assert actual["integer"]["sum"] is None
            assert actual["integer"]["min"] is None
            assert actual["integer"]["max"] is None
            assert actual["integer"]["mean"] is None

            assert np.isnan(actual["float"]["sum"])
            assert np.isnan(actual["float"]["min"])
            assert np.isnan(actual["float"]["max"])
            assert np.isnan(actual["float"]["mean"])

    def test_multiple_attrs(self):
        path = self.path("test_multiple_attrs")
        dom = tiledb.Domain(tiledb.Dim(name="d", domain=(0, 9), dtype=np.int32))
        attrs = [
            tiledb.Attr(name="integer", dtype=int),
            tiledb.Attr(name="float", dtype=float),
            tiledb.Attr(name="string", dtype=str),
        ]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=True)
        tiledb.Array.create(path, schema)

        with tiledb.open(path, "w") as A:
            A[np.arange(0, 10)] = {
                "integer": np.random.randint(1, 10, size=10),
                "float": np.random.randint(1, 10, size=10),
                "string": np.random.randint(1, 10, size=10).astype(str),
            }

        with tiledb.open(path, "r") as A:
            actual = A.query()[:]
            agg = A.query().agg

            assert agg({"string": "count"})[:] == len(actual["string"])
            invalid_aggregates = ("sum", "min", "max", "mean")
            for invalid_agg in invalid_aggregates:
                with pytest.raises(tiledb.TileDBError):
                    agg({"string": invalid_agg})[:]

            result = agg("count")[:]
            assert result["integer"]["count"] == len(actual["integer"])
            assert result["float"]["count"] == len(actual["float"])
            assert result["string"]["count"] == len(actual["string"])

            with pytest.raises(tiledb.TileDBError):
                agg("sum")[:]

            result = agg({"integer": "sum", "float": "sum"})[:]
            assert "string" not in result
            assert result["integer"]["sum"] == sum(actual["integer"])
            assert result["float"]["sum"] == sum(actual["float"])

            result = agg(
                {
                    "string": ("count",),
                    "integer": "sum",
                    "float": ["max", "min", "sum", "mean"],
                }
            )[:]
            assert result["string"]["count"] == len(actual["string"])
            assert result["integer"]["sum"] == sum(actual["integer"])
            assert result["float"]["max"] == max(actual["float"])
            assert result["float"]["min"] == min(actual["float"])
            assert result["float"]["sum"] == sum(actual["float"])
            assert result["float"]["mean"] == sum(actual["float"]) / len(
                actual["float"]
            )



# time_traveling.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# When run, this program will create a simple sparse array, write some data
# to it at specified timestamps, and read the entire array data back.
#

import numpy as np

import tiledb

# Name of the array to create.
array_name = "time_traveling"


def create_array():
    dom = tiledb.Domain(tiledb.Dim(domain=(0, 0), tile=1, dtype=np.int64))
    att = tiledb.Attr(name="num", dtype=np.int64)
    schema = tiledb.ArraySchema(sparse=True, domain=dom, attrs=(att,))
    tiledb.SparseArray.create(array_name, schema)


def write_array():
    # Open the array and write to it.
    for timestamp in range(1, 4):
        with tiledb.open(array_name, timestamp=timestamp, mode="w") as T:
            T[0] = timestamp


def read_array():
    # Open the array and read from it.
    with tiledb.open(array_name, mode="r") as T:
        print(T[:]["num"])

    with tiledb.open(array_name, mode="r", timestamp=(1, 2)) as T:
        print(T[:]["num"])

    with tiledb.open(array_name, mode="r", timestamp=(2, 3)) as T:
        print(T[:]["num"])

    with tiledb.open(array_name, mode="r", timestamp=1) as T:
        print(T[:]["num"])

    with tiledb.open(array_name, mode="r", timestamp=(1, None)) as T:
        print(T[:]["num"])


if tiledb.object_type(array_name) != "array":
    create_array()
    write_array()

read_array()



import ctypes
import os
import sys
import warnings

if os.name == "posix":
    if sys.platform == "darwin":
        lib_name = "libtiledb.dylib"
    else:
        lib_name = "libtiledb.so"
else:
    lib_name = "tiledb"

import numpy as np

# TODO: get rid of this - It is currently used for unified numpy printing accross numpy versions
np.set_printoptions(
    legacy="1.21" if np.lib.NumpyVersion(np.__version__) >= "1.22.0" else False
)
del np

from tiledb.libtiledb import version as libtiledb_version

if libtiledb_version()[0] == 2 and libtiledb_version()[1] >= 26:
    from .current_domain import CurrentDomain
    from .ndrectangle import NDRectangle

del libtiledb_version  # no longer needed

from .array import Array
from .array_schema import ArraySchema
from .attribute import Attr
from .consolidation_plan import ConsolidationPlan
from .ctx import Config, Ctx, default_ctx, scope_ctx
from .dataframe_ import from_csv, from_pandas, open_dataframe
from .dense_array import DenseArrayImpl
from .dimension import Dim
from .dimension_label import DimLabel
from .dimension_label_schema import DimLabelSchema
from .domain import Domain
from .enumeration import Enumeration
from .filestore import Filestore
from .filter import (
    BitShuffleFilter,
    BitWidthReductionFilter,
    ByteShuffleFilter,
    Bzip2Filter,
    ChecksumMD5Filter,
    ChecksumSHA256Filter,
    CompressionFilter,
    DeltaFilter,
    DictionaryFilter,
    DoubleDeltaFilter,
    Filter,
    FilterList,
    FloatScaleFilter,
    GzipFilter,
    LZ4Filter,
    NoOpFilter,
    PositiveDeltaFilter,
    RleFilter,
    WebpFilter,
    XORFilter,
    ZstdFilter,
)
from .fragment import (
    FragmentInfo,
    FragmentInfoList,
    copy_fragments_to_existing_array,
    create_array_from_fragments,
)
from .group import Group
from .highlevel import (
    array_exists,
    array_fragments,
    as_built,
    consolidate,
    empty_like,
    from_numpy,
    ls,
    move,
    object_type,
    open,
    remove,
    save,
    schema_like,
    vacuum,
    walk,
)
from .libtiledb import TileDBError
from .multirange_indexing import EmptyRange
from .object import Object
from .parquet_ import from_parquet
from .query import Query
from .query_condition import QueryCondition
from .schema_evolution import ArraySchemaEvolution
from .sparse_array import SparseArrayImpl
from .stats import (
    stats_disable,
    stats_dump,
    stats_enable,
    stats_reset,
)
from .subarray import Subarray
from .version_helper import version
from .vfs import VFS, FileIO

__version__ = version.version
group_create = Group.create

# Create a proxy object to wrap libtiledb and provide a `cc` alias
class CCProxy:
    def __init__(self, module):
        self._module = module

    def __getattr__(self, name):
        warnings.warn(
            "`tiledb.cc` is deprecated. Please use `tiledb.libtiledb` instead.",
        )
        return getattr(self._module, name)

    def __repr__(self):
        warnings.warn(
            "`tiledb.cc` is deprecated. Please use `tiledb.libtiledb` instead.",
        )
        return self._module.__repr__()


cc = CCProxy(libtiledb)
sys.modules["tiledb.cc"] = cc
cc = cc
del CCProxy

# Note: we use a modified namespace packaging to allow continuity of existing TileDB-Py imports.
#       Therefore, 'tiledb/__init__.py' must *only* exist in this package.
#       Furthermore, in sub-packages, the `find_packages` helper will not work at the
#       root directory due to lack of 'tiledb/__init__.py'. Sub-package 'setup.py' scripts
#       must declare constituents accordingly, such as by running 'find_packages' on a sub-directory
#       and applying prefixes accordingly.
#   1) https://packaging.python.org/guides/packaging-namespace-packages/#native-namespace-packages
#   2) https://stackoverflow.com/a/53486554
#
# Note: 'pip -e' in particular will not work without this declaration:
__path__ = __import__("pkgutil").extend_path(__path__, __name__)

# If tiledb.cloud is installed, add CloudArray methods to TileDB arrays
try:
    from tiledb.cloud.cloudarray import CloudArray
except ImportError:

    class DenseArray(DenseArrayImpl):
        pass

    class SparseArray(SparseArrayImpl):
        pass

else:

    class DenseArray(DenseArrayImpl, CloudArray):
        pass

    class SparseArray(SparseArrayImpl, CloudArray):
        pass

    del CloudArray



import io
import os
from types import TracebackType
from typing import Callable, List, Optional, Type, Union

import numpy as np

import tiledb
import tiledb.libtiledb as lt

from .ctx import Config, Ctx, default_ctx

_AnyPath = Union[str, bytes, os.PathLike]


class VFS(lt.VFS):
    """TileDB VFS class

    Encapsulates the TileDB VFS module instance with a specific configuration (config).

    :param tiledb.Ctx ctx: The TileDB Context
    :param config: Override `ctx` VFS configurations with updated values in config.
    :type config: tiledb.Config or dict

    """

    def __init__(self, config: Union[Config, dict] = None, ctx: Optional[Ctx] = None):
        self.ctx = ctx or default_ctx()

        if config:
            if isinstance(config, Config):
                config = config.dict()
            else:
                try:
                    config = dict(config)
                except Exception:
                    raise ValueError("`config` argument must be of type Config or dict")

            # Convert all values to strings
            self.config_dict = {k: str(v) for k, v in config.items()}

            ccfg = tiledb.Config(self.config_dict)
            super().__init__(self.ctx, ccfg)
        else:
            super().__init__(self.ctx)

    def ctx(self) -> Ctx:
        """
        :rtype: tiledb.Ctx
        :return: context associated with the VFS object
        """
        return self._ctx

    def config(self) -> Config:
        """
        :rtype: tiledb.Config
        :return: config associated with the VFS object
        """
        return self._config

    def open(self, uri: _AnyPath, mode: str = "rb"):
        """Opens a VFS file resource for reading / writing / appends at URI.

        If the file did not exist upon opening, a new file is created.

        :param str uri: URI of VFS file resource
        :param mode str: 'rb' for opening the file to read, 'wb' to write, 'ab' to append
        :rtype: FileHandle
        :return: TileDB FileIO
        :raises TypeError: cannot convert `uri` to unicode string
        :raises ValueError: invalid mode
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return FileIO(self, uri, mode)

    def close(self, file: lt.FileHandle):
        """Closes a VFS FileHandle object.

        :param FileIO file: An opened VFS FileIO
        :rtype: FileIO
        :return: closed VFS FileHandle
        :raises: :py:exc:`tiledb.TileDBError`

        """
        file.close()
        return file

    def write(self, file: lt.FileHandle, buff: Union[str, bytes]):
        """Writes buffer to opened VFS FileHandle.

        :param FileHandle file: An opened VFS FileHandle in 'w' mode
        :param buff: a Python object that supports the byte buffer protocol
        :raises TypeError: cannot convert buff to bytes
        :raises: :py:exc:`tiledb.TileDBError`

        """
        if isinstance(file, FileIO):
            raise lt.TileDBError(
                "`tiledb.VFS().open` now returns a FileIO object. Use "
                "`FileIO.write`. This message will be removed in 0.21.0.",
            )
        if isinstance(buff, str):
            buff = buff.encode()
        file.write(buff)

    def read(self, file: lt.FileHandle, offset: int, nbytes: int) -> bytes:
        """Read nbytes from an opened VFS FileHandle at a given offset.

        :param FileHandle file: An opened VFS FileHandle in 'r' mode
        :param int offset: offset position in bytes to read from
        :param int nbytes: number of bytes to read
        :rtype: :py:func:`bytes`
        :return: read bytes
        :raises: :py:exc:`tiledb.TileDBError`
        """
        if nbytes == 0:
            return b""

        return file.read(offset, nbytes)

    def supports(self, scheme: str) -> bool:
        """Returns true if the given URI scheme (storage backend) is supported.

        :param str scheme: scheme component of a VFS resource URI (ex. 'file' / 'hdfs' / 's3')
        :rtype: bool
        :return: True if the linked libtiledb version supports the storage backend, False otherwise
        :raises ValueError: VFS storage backend is not supported

        """
        if scheme == "file":
            return True

        scheme_to_fs_type = {
            "s3": lt.FileSystem.S3,
            "azure": lt.FileSystem.AZURE,
            "gcs": lt.FileSystem.GCS,
            "hdfs": lt.FileSystem.HDFS,
        }

        if scheme not in scheme_to_fs_type:
            raise ValueError(f"Unsupported VFS scheme '{scheme}://'")

        return self._ctx.is_supported_fs(scheme_to_fs_type[scheme])

    def create_bucket(self, uri: _AnyPath):
        """Creates an object store bucket with the input URI.

        :param str uri: Input URI of the bucket

        """
        return self._create_bucket(_to_path_str(uri))

    def remove_bucket(self, uri: _AnyPath):
        """Deletes an object store bucket with the input URI.

        :param str uri: Input URI of the bucket

        """
        return self._remove_bucket(_to_path_str(uri))

    def is_bucket(self, uri: _AnyPath) -> bool:
        """
        :param str uri: Input URI of the bucket
        :rtype: bool
        :return: True if an object store bucket with the input URI exists, False otherwise

        """
        return self._is_bucket(_to_path_str(uri))

    def empty_bucket(self, uri: _AnyPath):
        """Empty an object store bucket.

        :param str uri: Input URI of the bucket

        """
        return self._empty_bucket(_to_path_str(uri))

    def is_empty_bucket(self, uri: _AnyPath) -> bool:
        """
        :param str uri: Input URI of the bucket
        :rtype: bool
        :return: True if an object store bucket is empty, False otherwise

        """
        return self._is_empty_bucket(_to_path_str(uri))

    def create_dir(self, uri: _AnyPath):
        """Create a directory at the specified input URI.

        :param str uri: Input URI of the directory

        """
        return self._create_dir(_to_path_str(uri))

    def is_dir(self, uri: _AnyPath) -> bool:
        """
        :param str uri: Input URI of the directory
        :rtype: bool
        :return: True if a directory with the input URI exists, False otherwise

        """
        return self._is_dir(_to_path_str(uri))

    def remove_dir(self, uri: _AnyPath):
        """Removes a directory (recursively) with the input URI.

        :param str uri: Input URI of the directory

        """
        return self._remove_dir(_to_path_str(uri))

    def dir_size(self, uri: _AnyPath) -> int:
        """
        :param str uri: Input URI of the directory
        :rtype: int
        :return: The size of a directory with the input URI

        """
        return self._dir_size(_to_path_str(uri))

    def move_dir(self, old_uri: _AnyPath, new_uri: _AnyPath):
        """Renames a TileDB directory from an old URI to a new URI.

        :param str old_uri: Input of the old directory URI
        :param str new_uri: Input of the new directory URI

        """
        return self._move_dir(_to_path_str(old_uri), _to_path_str(new_uri))

    def copy_dir(self, old_uri: _AnyPath, new_uri: _AnyPath):
        """Copies a TileDB directory from an old URI to a new URI.

        :param str old_uri: Input of the old directory URI
        :param str new_uri: Input of the new directory URI

        """
        return self._copy_dir(_to_path_str(old_uri), _to_path_str(new_uri))

    def is_file(self, uri: _AnyPath) -> bool:
        """
        :param str uri: Input URI of the file
        :rtype: bool
        :return: True if a file with the input URI exists, False otherwise

        """
        return self._is_file(_to_path_str(uri))

    def remove_file(self, uri: _AnyPath):
        """Removes a file with the input URI.

        :param str uri: Input URI of the file

        """
        return self._remove_file(_to_path_str(uri))

    def file_size(self, uri: _AnyPath) -> int:
        """
        :param str uri: Input URI of the file
        :rtype: int
        :return: The size of a file with the input URI

        """
        return self._file_size(_to_path_str(uri))

    def move_file(self, old_uri: _AnyPath, new_uri: _AnyPath):
        """Renames a TileDB file from an old URI to a new URI.

        :param str old_uri: Input of the old file URI
        :param str new_uri: Input of the new file URI

        """
        return self._move_file(_to_path_str(old_uri), _to_path_str(new_uri))

    def copy_file(self, old_uri: _AnyPath, new_uri: _AnyPath):
        """Copies a TileDB file from an old URI to a new URI.

        :param str old_uri: Input of the old file URI
        :param str new_uri: Input of the new file URI

        """
        return self._copy_file(_to_path_str(old_uri), _to_path_str(new_uri))

    def ls(self, uri: _AnyPath, recursive: bool = False) -> List[str]:
        """Retrieves the children in directory `uri`. This function is
        non-recursive, i.e., it focuses in one level below `uri`.

        :param str uri: Input URI of the directory
        :param bool recursive: If True, recursively list all children in the directory
        :rtype: List[str]
        :return: The children in directory `uri`

        """
        if recursive:
            return VFS._ls_recursive(self, _to_path_str(uri), None)

        return self._ls(_to_path_str(uri))

    def ls_recursive(
        self, uri: _AnyPath, callback: Optional[Callable[[str, int], bool]] = None
    ):
        """Recursively lists objects at the input URI, invoking the provided callback
        on each entry gathered. The callback is passed the data pointer provided
        on each invocation and is responsible for writing the collected results
        into this structure. If the callback returns True, the walk will continue.
        If False, the walk will stop. If an error is thrown, the walk will stop and
        the error will be propagated to the caller using std::throw_with_nested.

        Currently only local, S3, Azure and GCS are supported.

        :param str uri: Input URI of the directory
        :param callback: Callback function to invoke on each entry

        """
        return VFS._ls_recursive(self, _to_path_str(uri), callback)

    def touch(self, uri: _AnyPath):
        """Touches a file with the input URI, i.e., creates a new empty file.

        :param str uri: Input URI of the file

        """
        return self._touch(_to_path_str(uri))

    # Aliases for compatibility with fsspec's AbstractFileSystem.
    isdir = is_dir
    isfile = is_file
    size = file_size

    # pickling support
    def __getstate__(self):
        # self.config_dict might not exist. In that case use the config from ctx.
        if hasattr(self, "config_dict"):
            config_dict = self.config_dict
        else:
            config_dict = self.config().dict()
        return (config_dict,)

    def __setstate__(self, state):
        config_dict = state[0]
        config = Config(params=config_dict)
        ctx = Ctx(config)
        self.__init__(config=config, ctx=ctx)


class FileIO(io.RawIOBase):
    """TileDB FileIO class that encapsulates files opened by tiledb.VFS. The file
    operations are meant to mimic Python's built-in file I/O methods."""

    def __init__(self, vfs: VFS, uri: _AnyPath, mode: str = "rb"):
        uri = _to_path_str(uri)
        self._vfs = vfs

        str_to_vfs_mode = {
            "rb": lt.VFSMode.READ,
            "wb": lt.VFSMode.WRITE,
            "ab": lt.VFSMode.APPEND,
        }
        if mode not in str_to_vfs_mode:
            raise ValueError(f"invalid mode {mode}")

        self._mode = mode
        self._offset = 0
        self._nbytes = 0

        if self._mode == "rb":
            try:
                self._nbytes = vfs.file_size(uri)
            except Exception as e:
                raise lt.TileDBError(f"URI {uri!r} is not a valid file") from e

        self._fh = lt.FileHandle(
            self._vfs._ctx, self._vfs, uri, str_to_vfs_mode[self._mode]
        )

    def __len__(self):
        """
        :rtype: int
        :return: Number of bytes in file

        """
        return self._nbytes

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        self._fh._close()

    @property
    def mode(self) -> str:
        """
        :rtype: str
        :return: Whether the file is in read mode ("rb"), write mode ("wb"), or append mode ("ab")

        """
        return self._mode

    def readable(self) -> bool:
        """
        :rtype: bool
        :return: True if the file is readable (ie. "rb" mode), otherwise False

        """
        return self._mode == "rb"

    def writable(self) -> bool:
        """
        :rtype: bool
        :return: True if the file is writable (ie. "wb" or "ab" mode), otherwise False

        """
        return self._mode != "rb"

    @property
    def closed(self) -> bool:
        """
        :rtype: bool
        :return: True if the file is closed, otherwise False

        """
        return self._fh._closed

    def seekable(self):
        """
        All tiledb.FileIO objects are seekable.

        :rtype: bool
        :return: True

        """
        return True

    def flush(self):
        """
        Force the data to be written to the file.
        """
        self._fh._flush()

    def seek(self, offset: int, whence: int = 0):
        """
        :param int offset: Byte position to set the file pointer
        :param int whence: Reference point. A whence value of 0 measures from the beginning of the file, 1 uses the current file position, and 2 uses the end of the file as the reference point. whence can be omitted and defaults to 0.
        """
        if not np.issubdtype(type(offset), np.integer):
            raise TypeError(
                f"Offset must be an integer or None (got type {type(offset)})"
            )
        if whence == 0:
            if offset < 0:
                raise ValueError(
                    "offset must be a positive or zero value when SEEK_SET"
                )
            self._offset = offset
        elif whence == 1:
            self._offset += offset
        elif whence == 2:
            self._offset = self._nbytes + offset
        else:
            raise ValueError("whence must be equal to SEEK_SET, SEEK_START, SEEK_END")
        if self._offset < 0:
            self._offset = 0
        elif self._offset > self._nbytes:
            self._offset = self._nbytes

        return self._offset

    def tell(self) -> int:
        """
        :rtype: int
        :return: The current position in the file represented as number of bytes

        """
        return self._offset

    def read(self, size: int = -1) -> bytes:
        """
        Read the file from the current pointer position.

        :param int size: Number of bytes to read. By default, size is set to -1 which will read until the end of the file.
        :rtype: bytes
        :return: The bytes in the file

        """
        if not np.issubdtype(type(size), np.integer):
            raise TypeError(f"size must be an integer or None (got type {type(size)})")
        if not self.readable():
            raise IOError("Cannot read from write-only FileIO handle")
        if self.closed:
            raise IOError("Cannot read from closed FileIO handle")

        nbytes_left = self._nbytes - self._offset
        nbytes = nbytes_left if size < 0 or size > nbytes_left else size
        if nbytes == 0:
            return b""

        buff = self._fh._read(self._offset, nbytes)
        self._offset += nbytes
        return buff

    def write(self, buff: bytes):
        """
        :param bytes buff: Write the given bytes to the file

        """
        if not self.writable():
            raise IOError("cannot write to read-only FileIO handle")
        if isinstance(buff, str):
            buff = buff.encode()
        nbytes = len(buff)
        self._fh._write(buff)
        self._nbytes += nbytes
        self._offset += nbytes
        return nbytes

    def readinto(self, buff: Union[bytes, bytearray, memoryview]) -> int:
        """
        Read bytes into a pre-allocated, writable, bytes-like object, and return the number of bytes read.
        :param buff bytes | bytearray | memoryview: A pre-allocated, writable object that supports the byte buffer protocol
        :rtype: int
        :return: The number of bytes read
        """
        buff = memoryview(buff).cast("b")
        size = len(buff)
        if not self.readable():
            raise IOError("Cannot read from write-only FileIO handle")
        if self.closed:
            raise IOError("Cannot read from closed FileIO handle")

        nbytes_left = self._nbytes - self._offset
        nbytes = nbytes_left if size > nbytes_left else size
        if nbytes == 0:
            return None

        buff_temp = self._fh._read(self._offset, nbytes)
        self._offset += nbytes
        buff[: len(buff_temp)] = memoryview(buff_temp).cast("b")
        return len(buff_temp)

    def readinto1(self, b):
        return self.readinto(b)


def _to_path_str(pth: _AnyPath) -> Union[str, bytes]:
    if isinstance(pth, (str, bytes)):
        return pth
    try:
        return pth.__fspath__()
    except AttributeError as ae:
        raise TypeError(
            "VFS paths must be strings, bytes, or os.PathLike objects"
        ) from ae



# multirange_indexing.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#  https://docs.tiledb.com/main/how-to/arrays/reading-arrays/multi-range-subarrays
#
# When run, this program will create a simple 2D dense array with two
# attributes, write some data to it, and read a slice of the data back on
# (i) both attributes, and (ii) subselecting on only one of the attributes.
#


import numpy as np

import tiledb

# Name of the array to create.
array_name = "multi_range"


def create_array():
    # Check if the array already exists.
    if tiledb.object_type(array_name) == "array":
        return

    dom = tiledb.Domain(
        tiledb.Dim(name="x", domain=(1, 20), tile=4, dtype=np.int64),
        tiledb.Dim(name="y", domain=(1, 20), tile=4, dtype=np.int64),
    )

    # Add a single "a" float attribute
    schema = tiledb.ArraySchema(
        domain=dom, sparse=False, attrs=[tiledb.Attr(name="a", dtype=np.float64)]
    )

    # Create the (empty) array on disk.
    tiledb.DenseArray.create(array_name, schema)


def write_array():
    # Open the array and write to it.
    with tiledb.DenseArray(array_name, mode="w") as A:
        data_a = np.arange(400).reshape(20, 20)
        A[:, :] = {"a": data_a}


def read_array():
    # Open the array and read from it.
    with tiledb.DenseArray(array_name, mode="r") as A:
        # Slice only rows: (1,3) inclusive, and 5
        #            cols: 2, 5, 7
        data = A.multi_index[[(1, 3), 5], [2, 5, 7]]
        print("Reading attribute 'a', [ [1:3, 5], [2,5,7] ]")
        a = data["a"]
        print(a)


create_array()
write_array()
read_array()



import ast
import subprocess
import sys

from packaging.version import Version


def tiledb_cloud_eagerly_imports_pandas() -> bool:
    try:
        import pandas

        import tiledb.cloud
    except ImportError:
        # Can't import something that's not installed.
        return False
    if Version(tiledb.cloud.__version__) < Version("0.10.21"):
        # Old versions of tiledb-cloud will import Pandas eagerly.
        return True
    if Version(pandas.__version__) < Version("1.5"):
        # If an old version of Pandas is installed, tiledb-cloud needs to
        # import it eagerly to patch it.
        return True
    return False


def test_dont_import_pandas() -> None:
    """Verifies that when we import TileDB, we don't import Pandas eagerly."""

    # We import tiledb.cloud within tiledb-py, if available, in order to hook
    # Array.apply and other functionality.  If the version of tiledb-cloud
    # we have installed would import Pandas eagerly on its own, we need to
    # suppress its importation.
    suppress_cloud = (
        "sys.modules['tiledb.cloud'] = None;"
        if tiledb_cloud_eagerly_imports_pandas()
        else ""
    )
    # Get a list of all modules from a completely fresh interpreter.
    all_mods_str = subprocess.check_output(
        (
            sys.executable,
            "-c",
            f"import sys; {suppress_cloud} import tiledb; print(list(sys.modules))",
        )
    )
    all_mods = ast.literal_eval(all_mods_str.decode())
    assert "pandas" not in all_mods



# incomplete_iteration.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#   https://docs.tiledb.com/main/how-to/arrays/reading-arrays/incomplete-queries
#
# When run, this program will create a 1D dense array, write some data
# to it, and read slices back by iteration over incomplete queries.
#


import numpy as np

import tiledb


def check_dataframe_deps():
    pd_error = """Pandas version >= 1.0 and < 3.0 required for dataframe functionality.
                  Please `pip install pandas>=1.0,<3.0` to proceed."""

    try:
        import pandas as pd
    except ImportError:
        raise Exception(pd_error)

    from packaging.version import Version

    if Version(pd.__version__) < Version("1.0") or Version(pd.__version__) >= Version(
        "3.0.0.dev0"
    ):
        raise Exception(pd_error)


# Name of the array to create.
array_name = "incomplete_iteration"


def create_array():
    # The array will be 100 cells with dimensions "x".
    dom = tiledb.Domain(tiledb.Dim(name="x", domain=(0, 99), tile=100, dtype=np.int64))

    # The array will be dense with a single string typed attribute "a"
    schema = tiledb.ArraySchema(
        domain=dom, sparse=True, attrs=[tiledb.Attr(name="a", dtype=str)]
    )

    # Create the (empty) array on disk.
    tiledb.SparseArray.create(array_name, schema)


def write_array():
    # Open the array and write to it.
    with tiledb.open(array_name, mode="w") as A:
        extent = A.schema.domain.dim("x").domain
        ncells = extent[1] - extent[0] + 1

        # Data is the Latin alphabet with varying repeat lengths
        data = [chr(i % 26 + 97) * (i % 52) for i in range(ncells)]

        # Coords are the dimension range
        coords = np.arange(extent[0], extent[1] + 1)

        A[coords] = data


def read_array_iterated():
    # in order to force iteration, restrict the buffer sizes
    # this setting gives 5 iterations for the example data
    init_buffer_bytes = 800
    cfg = tiledb.Config(
        {
            "py.init_buffer_bytes": init_buffer_bytes,
            "py.exact_init_buffer_bytes": "true",
        }
    )

    with tiledb.open(array_name, config=cfg) as A:
        # iterate over results as a dataframe
        iterable = A.query(return_incomplete=True).df[:]

        for i, result in enumerate(iterable):
            print(f"--- result {i} is a '{type(result)}' with size {len(result)}")
            print(result)
            print("---")

    print(f"Query completed after {i} iterations")


check_dataframe_deps()
create_array()
write_array()
read_array_iterated()



import os
import time

import numpy as np
from numpy.testing import assert_array_equal

import tiledb.libtiledb as lt


def test_group_metadata(tmp_path):
    int_data = np.array([1, 2, 3])
    flt_data = np.array([1.5, 2.5, 3.5])

    ctx = lt.Context()

    grp_path = os.path.join(tmp_path, "test_groups_metadata")
    lt.Group._create(ctx, grp_path)
    grp = lt.Group(ctx, grp_path, lt.QueryType.WRITE)
    grp._put_metadata("int", int_data)
    grp._put_metadata("flt", flt_data)
    grp._close()

    grp._open(lt.QueryType.READ)
    assert grp._metadata_num() == 2
    assert grp._has_metadata("int")
    assert_array_equal(grp._get_metadata("int", False)[0], int_data)
    assert grp._has_metadata("flt")
    assert_array_equal(grp._get_metadata("flt", False)[0], flt_data)
    grp._close()

    grp._open(lt.QueryType.WRITE)
    grp._delete_metadata("int")
    grp._close()

    grp = lt.Group(ctx, grp_path, lt.QueryType.READ)
    assert grp._metadata_num() == 1
    assert not grp._has_metadata("int")
    grp._close()


def test_group_members(tmp_path):
    ctx = lt.Context()

    grp_path = os.path.join(tmp_path, "test_group_metadata")
    lt.Group._create(ctx, grp_path)
    grp = lt.Group(ctx, grp_path, lt.QueryType.WRITE)

    subgrp_path = os.path.join(tmp_path, "test_group_0")
    lt.Group._create(ctx, subgrp_path)
    grp._add(subgrp_path)
    grp._close()

    grp._open(lt.QueryType.READ)
    assert grp._member_count() == 1
    member = grp._member(0)
    assert os.path.basename(member._uri) == os.path.basename(subgrp_path)
    assert member._type == lt.ObjectType.GROUP
    grp._close()

    grp._open(lt.QueryType.WRITE)
    grp._remove(subgrp_path)
    grp._close()

    grp = lt.Group(ctx, grp_path, lt.QueryType.READ)
    assert grp._member_count() == 0
    grp._close()



# query_condition_datetime.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2021 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

# This example creates an array with one datetime-typed attribute,
# writes sample data to the array, and then prints out a filtered
# dataframe using the TileDB QueryCondition feature to select on
# either equality or ranges of the generated attribute values.

import numpy as np
import pandas as pd

import tiledb

uri = "query_condition_datetime"

data = pd.DataFrame(
    np.sort(np.random.randint(438923600, 243892360000, 20, dtype=np.int64)).astype(
        "M8[ns]"
    ),
    columns=["dates"],
)
data.sort_values(by="dates")

tiledb.from_pandas(
    uri,
    data,
    column_types={"dates": "datetime64[ns]"},
)

with tiledb.open(uri) as A:
    # filter by exact match with the fifth cell
    search_date = data["dates"][5].to_numpy().astype(np.int64)
    result = A.query(cond=f"dates == {search_date}").df[:]

    print()
    print("Attribute dates matching index 5:")
    print(result)

    # filter values between cell index 3 and 8
    d1 = data["dates"].iloc[3].to_numpy().astype(np.int64)
    d2 = data["dates"].iloc[8].to_numpy().astype(np.int64)
    result2 = A.query(cond=f"dates > {d1} and dates < {d2}").df[:]

    print()
    print("Attribute dates where 'dates[3] < val < dates[8]'")
    print(result2)



import tiledb.libtiledb as lt

from .ctx import Ctx, CtxMixin
from .domain import Domain
from .ndrectangle import NDRectangle


class CurrentDomain(CtxMixin, lt.CurrentDomain):
    """
    Represents a TileDB current domain.
    """

    def __init__(self, ctx: Ctx):
        """Class representing the current domain of a TileDB Array.

        :param ctx: A TileDB Context
        :raises tiledb.TileDBError:
        """
        super().__init__(ctx)

    @property
    def type(self):
        """The type of the current domain.

        :rtype: tiledb.CurrentDomainType
        """
        return self._type

    @property
    def is_empty(self):
        """Checks if the current domain is empty.

        :rtype: bool
        """
        return self._is_empty

    def set_ndrectangle(self, ndrect: NDRectangle):
        """Sets an N-dimensional rectangle representation on a current domain.

        :param ndrect: The N-dimensional rectangle to be used.
        :raises tiledb.TileDBError:
        """
        self._set_ndrectangle(ndrect)

    @property
    def ndrectangle(self):
        """Gets the N-dimensional rectangle associated with the current domain object.

        :rtype: NDRectangle
        :raises tiledb.TileDBError:
        """
        return NDRectangle.from_pybind11(self._ctx, self._ndrectangle())



import concurrent
import concurrent.futures
import json
import os
import subprocess
import sys

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import tiledb

from .common import DiskTestCase, has_pandas, has_pyarrow

pd = pytest.importorskip("pandas")
tm = pd._testing


class FixesTest(DiskTestCase):
    def test_sc50378_overflowerror_python_int_too_large_to_convert_to_c_long(self):
        uri = self.path(
            "test_sc50378_overflowerror_python_int_too_large_to_convert_to_c_long"
        )
        MAX_UINT64 = np.iinfo(np.uint64).max
        dim = tiledb.Dim(
            name="id",
            domain=(0, MAX_UINT64 - 1),
            dtype=np.dtype(np.uint64),
        )
        dom = tiledb.Domain(dim)
        text_attr = tiledb.Attr(name="text", dtype=np.dtype("U1"), var=True)
        attrs = [text_attr]
        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=True,
            allows_duplicates=False,
            attrs=attrs,
        )
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, "w") as A:
            external_ids = np.array([0, 100, MAX_UINT64 - 1], dtype=np.dtype(np.uint64))
            data = {"text": np.array(["foo", "bar", "baz"], dtype="<U3")}
            A[external_ids] = data

        array = tiledb.open(uri, "r", timestamp=None, config=None)
        array[0]["text"][0]
        array[100]["text"][0]
        # This used to fail
        array[MAX_UINT64 - 1]["text"][0]

    def test_ch7727_float32_dim_estimate_incorrect(self):
        # set max allocation: because windows won't overallocate
        with tiledb.scope_ctx({"py.alloc_max_bytes": 1024**2 * 100}):
            uri = self.path()
            dom = tiledb.Domain(tiledb.Dim("x", domain=(1, 100), dtype=np.float32))
            att = tiledb.Attr("", dtype=np.bytes_)
            schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=True)
            tiledb.Array.create(uri, schema)

            with tiledb.open(uri, mode="w") as T:
                T[50.4] = b"hello"

            with tiledb.open(uri, mode="r") as T:
                assert T[:][""] == b"hello"
                assert T[50.4][""] == b"hello"

    def test_ch8292(self):
        # test fix for ch8292
        # We need to ensure that py.alloc_max_bytes is *not* applied to
        # dense arrays. Dense arrays should have exact estimates based
        # on the ranges, so there should be no risk of over-estimates.
        # This test sets py.alloc_max_bytes to 1 less than the expected
        # result array size, and asserts that the allocated buffers match
        # the expected result size rather than py.alloc_max_bytes.
        uri = self.path()
        max_val = np.iinfo(np.uint8).max
        with tiledb.from_numpy(uri, np.uint8(range(max_val))):
            pass
        with tiledb.scope_ctx(
            {"py.init_buffer_bytes": 2 * 1024**2, "py.alloc_max_bytes": 1024**2}
        ) as ctx3:
            with tiledb.open(uri) as b:
                q = tiledb.main.PyQuery(ctx3, b, ("",), (), 0, False)
                q._return_incomplete = True
                subarray = tiledb.Subarray(b, ctx3)
                subarray.add_ranges([[(0, max_val - 1)]])
                q.set_subarray(subarray)
                q._allocate_buffers()
                buffers = list(*q._get_buffers().values())
                assert buffers[0].nbytes == max_val

    @pytest.mark.skipif(not has_pandas(), reason="pandas>=1.0,<3.0 not installed")
    def test_ch10282_concurrent_multi_index(self):
        """Test concurrent access to a single tiledb.Array using
        Array.multi_index and Array.df. We pass an array and slice
        into a function run by a set of futures, along with expected
        result; then assert that the result from TileDB matches the
        expectation.
        """

        def slice_array(a: tiledb.Array, indexer, selection, expected):
            """Helper function to slice a given tiledb.Array with an indexer
            and assert that the selection matches the expected result."""
            res = getattr(a, indexer)[selection][""]
            if indexer == "df":
                res = res.values

            assert_array_equal(res, expected)

        uri = self.path()

        data = np.random.rand(100)
        with tiledb.from_numpy(uri, data):
            pass

        futures = []
        with tiledb.open(uri) as A:
            with concurrent.futures.ThreadPoolExecutor(10) as executor:
                for indexer in ["multi_index", "df"]:  #
                    for end_idx in range(1, 100, 5):
                        sel = slice(0, end_idx)
                        expected = data[sel.start : sel.stop + 1]
                        futures.append(
                            executor.submit(slice_array, A, indexer, sel, expected)
                        )

                concurrent.futures.wait(futures)

            # Important: must get each result here or else assertion
            # failures or exceptions will disappear.
            list(map(lambda x: x.result(), futures))

    # skip, does not currently work, because we cannot force use
    # of the memory estimate
    @pytest.mark.skip
    def test_sc16301_arrow_extra_estimate_dense(self):
        """
        Test that dense query of array with var-length attribute completes
        in one try. We are currently adding an extra element to the offset
        estimate from libtiledb, in order to avoid an unnecessary pair of
        query resubmits when the offsets won't fit in the estimated buffer.
        """

        uri = self.path("test_sc16301_arrow_extra_estimate_dense")

        dim1 = tiledb.Dim(name="d1", dtype="int64", domain=(1, 3))
        att = tiledb.Attr(name="a1", dtype="<U0", var=True)

        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(dim1),
            attrs=(att,),
            sparse=False,
            allows_duplicates=False,
        )
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, "w") as A:
            A[:] = np.array(["aaa", "bb", "c"])

        with tiledb.open(uri) as A:
            tiledb.stats_enable()
            A[:]

            stats_dump_str = tiledb.stats_dump(print_out=False)
            if tiledb.libtiledb.version() >= (2, 27):
                assert """"Context.Query.Reader.loop_num": 1""" in stats_dump_str
            else:
                assert (
                    """"Context.StorageManager.Query.Reader.loop_num": 1"""
                    in stats_dump_str
                )
            tiledb.stats_disable()

    def test_sc58286_fix_stats_dump_return_value_broken(self):
        uri = self.path("test_sc58286_fix_stats_dump_return_value_broken")
        dim1 = tiledb.Dim(name="d1", dtype="int64", domain=(1, 3))
        att = tiledb.Attr(name="a1", dtype="<U0", var=True)

        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(dim1),
            attrs=(att,),
            sparse=False,
            allows_duplicates=False,
        )
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, "w") as A:
            A[:] = np.array(["aaa", "bb", "c"])

        with tiledb.open(uri) as A:
            tiledb.stats_enable()
            A[:]

            # check that the stats cannot be parsed as json
            stats = tiledb.stats_dump(print_out=False, json=False)
            assert isinstance(stats, str)
            with pytest.raises(json.decoder.JSONDecodeError):
                json.loads(stats)

            stats = tiledb.stats_dump(print_out=False, json=False, include_python=True)
            assert isinstance(stats, str)
            with pytest.raises(json.decoder.JSONDecodeError):
                json.loads(stats)

            # check that the stats can be parsed as json
            stats = tiledb.stats_dump(print_out=False, json=True)
            assert isinstance(stats, str)
            json.loads(stats)

            stats = tiledb.stats_dump(print_out=False, json=True, include_python=True)
            assert isinstance(stats, str)
            res = json.loads(stats)

            tiledb.stats_disable()

            # check that some fields are present in the json output and are of the correct type
            assert "counters" in res and isinstance(res["counters"], dict)
            assert "timers" in res and isinstance(res["timers"], dict)
            assert "python" in res and isinstance(res["python"], dict)

    def test_fix_stats_error_messages(self):
        # Test that stats_dump prints a user-friendly error message when stats are not enabled
        with pytest.raises(tiledb.TileDBError) as exc:
            tiledb.stats_dump()
        assert "Statistics are not enabled. Call tiledb.stats_enable() first." in str(
            exc.value
        )

    @pytest.mark.skipif(
        not has_pandas() and has_pyarrow(),
        reason="pandas>=1.0,<3.0 or pyarrow>=1.0 not installed",
    )
    def test_py1078_df_all_empty_strings(self):
        uri = self.path()
        df = pd.DataFrame(
            index=np.arange(10).astype(str),
            data={
                "A": list(str(i) for i in range(10)),
                "B": list("" for i in range(10)),
            },
        )

        tiledb.from_pandas(uri, df)
        with tiledb.open(uri) as arr:
            tm.assert_frame_equal(arr.df[:], df)

    @pytest.mark.skipif(
        tiledb.libtiledb.version() < (2, 14, 0),
        reason="SC-23287 fix not implemented until libtiledb 2.14",
    )
    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="TODO does not run on windows due to env passthrough",
    )
    def test_sc23827_aws_region(self):
        # Test for SC-23287
        # The expected behavior here for `vfs.s3.region` is:
        # - default to '' if no environment variables are set
        # - empty if AWS_REGION or AWS_DEFAULT_REGION is set (to any value)

        def get_config_with_env(env, key):
            python_exe = sys.executable
            cmd = """import tiledb; print(tiledb.Config()[\"{}\"])""".format(key)
            test_path = os.path.dirname(os.path.abspath(__file__))

            sp_output = subprocess.check_output(
                [python_exe, "-c", cmd], cwd=test_path, env=env
            )
            return sp_output.decode("UTF-8").strip()

        if tiledb.libtiledb.version() >= (2, 27, 0):
            assert get_config_with_env({}, "vfs.s3.region") == ""
        else:
            assert get_config_with_env({}, "vfs.s3.region") == "us-east-1"
        assert get_config_with_env({"AWS_DEFAULT_REGION": ""}, "vfs.s3.region") == ""
        assert get_config_with_env({"AWS_REGION": ""}, "vfs.s3.region") == ""

    @pytest.mark.skipif(not has_pandas(), reason="pandas>=1.0,<3.0 not installed")
    @pytest.mark.parametrize("is_sparse", [True, False])
    def test_sc1430_nonexisting_timestamp(self, is_sparse):
        path = self.path("nonexisting_timestamp")

        if is_sparse:
            tiledb.from_pandas(
                path, pd.DataFrame({"a": np.random.rand(4)}), sparse=True
            )

            with tiledb.open(path, timestamp=1) as A:
                assert pd.DataFrame.equals(
                    A.df[:]["a"], pd.Series([], dtype=np.float64)
                )
        else:
            with tiledb.from_numpy(path, np.random.rand(4)) as A:
                pass

            with tiledb.open(path, timestamp=1) as A:
                assert_array_equal(A[:], np.ones(4) * np.nan)

    def test_sc27374_hilbert_default_tile_order(self):
        import os
        import shutil

        import tiledb

        uri = "repro"
        if os.path.exists(uri):
            shutil.rmtree(uri)

        dom = tiledb.Domain(
            tiledb.Dim(
                name="var_id",
                domain=(None, None),
                dtype="ascii",
                filters=[tiledb.ZstdFilter(level=1)],
            ),
        )

        attrs = []

        sch = tiledb.ArraySchema(
            domain=dom,
            attrs=attrs,
            sparse=True,
            allows_duplicates=False,
            offsets_filters=[
                tiledb.DoubleDeltaFilter(),
                tiledb.BitWidthReductionFilter(),
                tiledb.ZstdFilter(),
            ],
            capacity=1000,
            cell_order="hilbert",
            tile_order=None,  # <-------------------- note
        )

        tiledb.Array.create(uri, sch)

        with tiledb.open(uri) as A:
            assert A.schema.cell_order == "hilbert"
            assert A.schema.tile_order is None

    def test_sc43221(self):
        # GroupMeta object did not have a representation test; repr failed due to non-existent attribute access in check.
        tiledb.Group.create("mem://tmp1")
        a = tiledb.Group("mem://tmp1")
        repr(a.meta)

    def test_sc56611(self):
        # test from_numpy with sparse argument set to True
        uri = self.path("test_sc56611")
        data = np.random.rand(10, 10)
        with pytest.raises(tiledb.libtiledb.TileDBError) as exc_info:
            tiledb.from_numpy(uri, data, sparse=True)
        assert str(exc_info.value) == "from_numpy only supports dense arrays"


class SOMA919Test(DiskTestCase):
    """
    ORIGINAL CONTEXT:
    https://github.com/single-cell-data/TileDB-SOMA/issues/919
    https://gist.github.com/atolopko-czi/26683305258a9f77a57ccc364916338f

    We've distilled @atolopko-czi's gist example using the TileDB-Py API directly.
    """

    def run_test(self):
        import tempfile

        import numpy as np

        import tiledb

        root_uri = tempfile.mkdtemp()
        group_ctx100 = tiledb.Ctx()
        timestamp = None

        # create the group and add a dummy subgroup "causes_bug"
        tiledb.Group.create(root_uri, ctx=group_ctx100)
        with tiledb.Group(root_uri, "w", ctx=group_ctx100) as expt:
            tiledb.Group.create(root_uri + "/causes_bug", ctx=group_ctx100)
            expt.add(name="causes_bug", uri=root_uri + "/causes_bug")

        # add an array to the group (in a separate write operation)
        with tiledb.Group(root_uri, mode="w", ctx=group_ctx100) as expt:
            df_path = os.path.join(root_uri, "df")
            tiledb.from_numpy(df_path, np.ones((100, 100)), timestamp=timestamp)
            expt.add(name="df", uri=df_path)

        # check our view of the group at current time;
        # (previously, "df" is sometimes missing (non-deterministic)
        with tiledb.Group(root_uri) as expt:
            assert "df" in expt

        # IMPORTANT: commenting out either line 29 or 32 (individually) makes df always visible.
        # That is, to invite the bug we must BOTH add the causes_bug sibling element AND then reopen
        # the group write handle to add df. The separate reopen (line 32) simulates
        # tiledbsoma.tdb_handles.Wrapper._flush_hack().

    @pytest.mark.skipif(
        tiledb.libtiledb.version() < (2, 15, 0),
        reason="SOMA919 fix implemented in libtiledb 2.15",
    )
    def test_soma919(self):
        N = 100
        fails = 0
        for i in range(N):
            try:
                self.run_test()
            except AssertionError:
                fails += 1
        if fails > 0:
            pytest.fail(f"SOMA919 test, failure rate {100*fails/N}%")



from numbers import Real
from typing import Dict, Optional, Sequence, Tuple, Union

import numpy as np

import tiledb.libtiledb as lt

from .array import Array
from .ctx import Ctx, CtxMixin, default_ctx

Scalar = Real
Range = Tuple[Scalar, Scalar]


class Subarray(CtxMixin, lt.Subarray):
    """
    Represents a TileDB subarray.
    """

    def __init__(
        self,
        array: Array,
        ctx: Ctx = None,
    ):
        """Class representing a subarray of a TileDB Array.

        :param array: tiledb.Array the subarray is defined on
        :param ctx: A TileDB Context
        """
        self._array = array
        super().__init__(
            ctx, lt.Array(ctx if ctx is not None else default_ctx(), array)
        )

    def add_dim_range(self, dim_idx: int, dim_range: Range):
        """Add a range to a dimension of the subarray.

        :param dim_idx: dimension index (int) of the dimension to add the range to
        :param range: tuple containing the inclusive range to query on
        :raises: :py:exc:`tiledb.TileDBError`
        """
        self._add_dim_range(dim_idx, dim_range)

    def add_label_range(self, label: str, label_range: Range):
        """Add a dimension label to a dimension of the subarray.

        :param label: name (str) of the label to add a range to
        :param range: tuple containing the inclusive range to query on
        :raises: :py:exc:`tiledb.TileDBError`
        """
        self._add_label_range(self._ctx, label, label_range)

    def add_ranges(
        self,
        dim_ranges: Optional[Sequence[Sequence[Range]]] = None,
        label_ranges: Optional[Dict[str, Sequence[Range]]] = None,
    ):
        """Add ranges to the subarray.

        :param dim_ranges: A sequence of a sequence of ranges to add for each dimension
            on the subarray. Each range may either be a tuple (inclusive range to query
            on) or  numpy.ndarray (series of point ranges).
        :param label_ranges: A dictionary of label name to a sequence of ranges for the
            dimension label on the subarray. Each range must be a tuple (inclusive range
            to query on).
        :raises: :py:exc:`tiledb.TileDBError`
        """
        if dim_ranges:
            if any(isinstance(r, np.ndarray) for r in dim_ranges):
                self._add_ranges_bulk(self._ctx, dim_ranges)
            else:
                self._add_ranges(self._ctx, dim_ranges)
        if label_ranges:
            self._add_label_ranges(self._ctx, label_ranges)

    def has_label_range(self, dim_idx):
        """Returns if dimension label ranges are set on the requested dimension.

        :param dim_idx: Index (int) of the dimension to check for labels.
        :rtype: int
        """
        return self._has_label_range(self._ctx, dim_idx)

    def num_dim_ranges(self, key: Union[int, str]) -> np.uint64:
        """Returns the number of ranges on a dimension.

        :param key: dimension index (int) or name (str)
        :rtype: np.uint64
        """
        return self._range_num(key)

    def num_label_ranges(self, label: str) -> np.uint64:
        """Returns the number of ranges on a dimension label.

        :param key: dimension label name
        :rtype: np.uint64
        """
        if not isinstance(label, str):
            raise TypeError(f"invalid type {type(label)} for label")
        return self._label_range_num(self._ctx, label)

    def shape(self) -> Optional[Sequence[tuple]]:
        """Returns the shape of dense data using this subarray and ``None`` for sparse
        arrays.

        Warning: This is the shape of time dimension ranges. If a label range is set
        and the dimension range has not been computed, it will return `0`.

        :rtype: tuple(int, ...)
        :raises: :py:exc:`tiledb.TileDBError`
        """
        if self._array.schema.sparse:
            return None
        shape = self._shape(self._ctx)
        return tuple(length for length in shape)



# vfs.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#  https://docs.tiledb.com/main/how-to/virtual-filesystem
#
# This program explores the various TileDB VFS tools.
#

import os
import struct

import tiledb


def path(p):
    return os.path.join(os.getcwd(), p)


def dirs_files():
    # Create TileDB VFS
    vfs = tiledb.VFS()

    # Create directory
    if not vfs.is_dir("dir_A"):
        vfs.create_dir(path("dir_A"))
        print("Created 'dir_A'")
    else:
        print("'dir_A' already exists")

    # Creating an (empty) file
    if not vfs.is_file("dir_A/file_A"):
        vfs.touch(path("dir_A/file_A"))
        print("Created empty file 'dir_A/file_A'")
    else:
        print("'dir_A/file_A' already exists")

    # Getting the file size
    print("Size of file 'dir_A/file_A': {}".format(vfs.file_size(path("dir_A/file_A"))))

    # Moving files (moving directories is similar)
    print("Moving file 'dir_A/file_A' to 'dir_A/file_B'")
    vfs.move_file(path("dir_A/file_A"), path("dir_A/file_B"))

    # Deleting files and directories
    print("Deleting 'dir_A/file_B' and 'dir_A'")
    vfs.remove_file(path("dir_A/file_B"))
    vfs.remove_dir(path("dir_A"))


def write():
    # Create TileDB VFS
    vfs = tiledb.VFS()

    # Write binary data
    with vfs.open("tiledb_vfs.bin", "wb") as f:
        f.write(struct.pack("f", 153.0))
        f.write("abcd".encode("utf-8"))

    # Write binary data again - this will overwrite the previous file
    with vfs.open("tiledb_vfs.bin", "wb") as f:
        f.write(struct.pack("f", 153.1))
        f.write("abcdef".encode("utf-8"))

    # Append binary data to existing file (this will NOT work on S3)
    with vfs.open("tiledb_vfs.bin", "ab") as f:
        f.write("ghijkl".encode("utf-8"))


def read():
    # Create TileDB VFS
    vfs = tiledb.VFS()

    # Read binary data
    with vfs.open("tiledb_vfs.bin", "rb") as f:
        # Read the first 4 bytes (bytes [0:4])
        f1 = struct.unpack("f", f.read(4))[0]

        # Read the next 8 bytes (bytes [4:12])
        s1 = bytes.decode(f.read(8), "utf-8")

        print(f"Binary read:\n{f1}\n{s1}")


dirs_files()
write()
read()



import os

import tiledb.libtiledb as lt


def test_dir(tmp_path):
    ctx = lt.Context()
    vfs = lt.VFS(ctx)

    path = os.path.join(tmp_path, "test_dir")

    vfs._create_dir(path)
    assert vfs._is_dir(path) is True
    assert vfs._dir_size(path) == 0
    vfs._remove_dir(path)
    assert vfs._is_dir(path) is False


def test_file_handle(tmp_path):
    ctx = lt.Context()
    vfs = lt.VFS(ctx)

    path = os.path.join(tmp_path, "test_file_handle")

    fh = lt.FileHandle(ctx, vfs, path, lt.VFSMode.WRITE)
    fh._write(b"Hello")

    fh = lt.FileHandle(ctx, vfs, path, lt.VFSMode.READ)
    assert fh._read(0, 5) == b"Hello"

    fh = lt.FileHandle(ctx, vfs, path, lt.VFSMode.APPEND)
    fh._write(b", world!")

    fh = lt.FileHandle(ctx, vfs, path, lt.VFSMode.READ)
    assert fh._read(0, 13) == b"Hello, world!"

    assert fh._closed is False
